# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Psychoacoustic NTF design (:mod:`pydsm.NTFdesign.psychoacoustic`)
=================================================================

This modules provides a psychoacoustic aware design method for the NTF
of ΔΣ modulators used in audio applications.

.. currentmodule:: pydsm.NTFdesign.psychoacoustic


Functions
---------

.. autosummary::
   :toctree: generated/

    dunn_optzeros            -- helper function for normalized NTF zeros
    dunn_optzeros_cplx       -- helper function for NTF zeros
    ntf_dunn                 -- NTF design a la Dunn
    ntf_fir_audio_weighting  -- FIR NTF design with psychoacustic weighting


Deprecated functions
--------------------

.. autosummary::
   :toctree: generated/

    synthesize_ntf_dunn                  -- alias of ntf_dunn
    synthesize_ntf_from_audio_weighting  -- alias of ntf_fir_audio_weighting
"""

from __future__ import division, print_function

import numpy as np
from .delsig import synthesizeNTF
from ..delsig import undbp as undbp
from .. import audio_weightings
from .weighting import ntf_fir_weighting
from warnings import warn
from ..exceptions import PyDsmDeprecationWarning

__all__ = ["dunn_optzeros", "dunn_optzeros_cplx", "synthesize_ntf_dunn",
           "synthesize_ntf_from_audio_weighting",
           "ntf_dunn", "ntf_fir_audio_weighting"]


def dunn_optzeros(n):
    """
    Helper function for synthesising psychoacoustically optimal modulators.

    Returns the (normalized) zeros which minimize the in-band noise power of
    a delta-sigma modulator's NTF after weighting by the F-weighting.

    Parameters
    ----------
    n : int
        the number of optimized zeros to return

    Returns
    -------
    zeros : ndarray of reals
        the zeros for the modulator as normalized angular frequencies.

    Notes
    -----
    The zeros are always located on the complex unit circle. In other words,
    the zeros are returned as frequencies. For homogeneity with DELSIG's
    ds_optzeros, the frequencies are normalized with respect to the signal
    bandwidth, that is fixed at 22.05 kHz.

    This function is the equivalent of ds_optzeros in DELSIG. The tabled
    zeros delivered by this function are from [1]_.  Note that this function
    does not return the values that are tabled in [1]_, but scales them
    by the reference audio bandwidth used in [1]_, namely 22.05 kHz.

    .. [1] Chris Dunn and Mark Sandler, "Psychoacoustically Optimal
       Sigma Delta Modulation," J. Audio Eng. Soc., Vol. 45, No. 4, pp.
       212 - 223 (1997 April)
    """
    # These are the optimal zero placements in kHz for a 22.05 kHz bandwidth
    # as found in the paper by Dunn and Sandler
    zero_freqs_unnorm = [[0.0],
                         [4.014, -4.014],
                         [0.0, 6.443, -6.443],
                         [3.590, -3.590, 11.954, -11.954],
                         [0.0, 4.308, -4.308, 12.959, -12.959],
                         [3.325, -3.325, 7.078, -7.078, 13.389, -13.389],
                         [0.0, 0.0, 4.017, -4.017, 10.471, -10.471, 13.842,
                          -13.842],
                         [2.933, -2.933, 5.167, -5.167, 12.012, -12.012,
                          14.381, -14.381]]
    if n > 8:
        raise ValueError('Optimized zeros for n>14 are not available.')
    return np.asarray(zero_freqs_unnorm[n-1])/22.05


def dunn_optzeros_cplx(order, osr):
    """
    Helper function for synthesising psychoacoustically optimal modulators.

    Returns the complex zeros which minimize the in-band noise power of
    a delta-sigma modulator's NTF after weighting by the F-weighting.
    The signal bandwidth is 22.05 kHz.

    Parameters
    ----------
    n : int
        the number of optimized zeros to return
    osr : real
        the oversampling ratio

    Returns
    -------
    zeros : ndarray of complex
        the zeros for the modulator as complex values

    See Also
    --------
    dunn_optzeros : Dunn's optimal zeros
    """
    w2 = dunn_optzeros(order)/osr*np.pi
    return np.exp(1j*w2)


def ntf_dunn(order=3, osr=64, H_inf=1.5):
    """
    Synthesizes an NTF for a DS audio modulator by Dunn's approach.

    The signal bandwidth is 22.05 kHz.

    Parameters
    ----------
    order : int, optional
        the order of the modulator, defaults to 3
    osr : float, optional
        the oversamping ratio (based on the actual signal bandwidth)
    H_inf : real, optional
        max allowed peak value of the NTF. Defaults to 1.5

    Returns
    -------
    ntf : tuple
        noise transfer function in zpk form.

    Warns
    -----
    As in DELSIG's synthesizeNTF

    Notes
    -----
    This is not exactly Dunn's method (but it should be equivalent or
    slightly better). In fact, to avoid re-implementing the pole selection
    algorithm, that in [1]_ is based on a Butterworth synthesis, here
    the pole selection logic used in DELSIG's synthesizeNTF is recycled.
    This should not make a big difference since DELSIG logic is anyway based
    on a pole positioning aimed at obtaining a maximally flat response of the
    NTF denominator in the signal band. This has the advantage of
    automatically controlling the peak gain of the NTF, as in the Lee
    criterion.

    Parameter H_inf is used to enforce the Lee stability criterion.

    .. [1] Chris Dunn and Mark Sandler, "Psychoacoustically Optimal
       Sigma Delta Modulation," J. Audio Eng. Soc., Vol. 45, No. 4, pp.
       212 - 223 (1997 April)

    See Also
    --------
    pydsm.delsig.synthesizeNTF : DELSIG's optimal NTF design strategy.
    """
    return synthesizeNTF(order, osr, dunn_optzeros_cplx(order, osr), H_inf, 0)


def ntf_fir_audio_weighting(
        order, osr,
        audio_weighting='f',
        audio_band=22.05E3,
        max_attn=120,
        H_inf=1.5,
        normalize="auto", **options):
    u"""Synthesize a FIR NTF based on an audio weighting function.

    The ΔΣ modulator NTF is designed after an audio weigthing function stating
    how loudly noise is perceived at the various frequencies.

    Parameters
    ----------
    order : int
        Delta sigma modulator order
    osr : float
        the oversampling ratio
    audio_weighting : callable or string
        audio weighting function. This is a function taking a frequency and
        expressing the weighting at that frequency in terms of acoustic power.
        Functions in the :mod:`pydsm.audio_weightings` module are suitable
        here. Note that the function argument frequency is a real frequency in
        Hz.
    audio_band : float, optional
        how large the audio bandwidth to consider. The signal band is from
        0 to audio_band Hz. Defaults to 22.05 kHz
    max_attn : float, optional
        clip very large attenuations to this value (in dB). This helps the
        convergenze of the optimization routine used to design the NTF.
        Defaults to 120 dB.
    H_inf : real, optional
        Max peak NTF gain, defaults to 1.5, used to enforce the Lee criterion
    normalize : string or real, optional
        Normalization to apply to the quadratic form used in the NTF
        selection. Defaults to 'auto' which means setting the top left entry
        in the matrix Q defining the quadratic form to 1.

    Returns
    -------
    ntf : ndarray
        FIR NTF in zpk form

    Other parameters
    ----------------
    show_progress : bool, optional
        provide extended output, default is True and can be updated by
        changing the function ``default_options`` attribute.
    cvxopt_opts : dictionary, optional
        A dictionary of options for the ``cvxopt`` optimizer
        Allowed options include:

        ``maxiters``
            Maximum number of iterations (defaults to 100)
        ``abstol``
            Absolute accuracy (defaults to 1e-7)
        ``reltol``
            Relative accuracy (defaults to 1e-6)
        ``feastol``
            Tolerance for feasibility conditions (defaults to 1e-6)

        Do not use other options since they could break ``CVXOPT`` in
        unexpected ways. Defaults can be set by changing the function
        ``default_options`` attribute.
    quad_opts : dictionary, optional
        Parameters to be passed to the :func:`scipy.integrate.quad` function
        used internally as an integrator. Allowed options are ``epsabs``,
        ``epsrel``, ``limit``, ``points``. Do not use other options since they
        could break the integrator in unexpected ways. Defaults can be set by
        changing the function ``default_options`` attribute.

    See Also
    --------
    scipy.integrate.quad : for the meaning of the integrator parameters.

    Notes
    -----
    For more information on the ``CVXOPT`` optimizer parameters, see
    the corresponding documentation.
    """
    if not callable(audio_weighting):
        if audio_weighting == 'f':
            audio_weighting = audio_weightings.f_weighting
        elif audio_weighting == 'a':
            audio_weighting = audio_weightings.a_weighting
        elif audio_weighting == 'b':
            audio_weighting = audio_weightings.b_weighting
        elif audio_weighting == 'c':
            audio_weighting = audio_weightings.c_weighting
        elif audio_weighting == 'd':
            audio_weighting = audio_weightings.d_weighting

    def w(f):
        ma = undbp(-max_attn)
        fx = f*audio_band*2*osr
        w = audio_weighting(fx) if fx <= audio_band else 0
        return max(w, ma)

    return ntf_fir_weighting(order, w, H_inf, normalize, **options)

ntf_fir_audio_weighting.default_options = \
    ntf_fir_weighting.default_options.copy()


# Following part is deprecated

def synthesize_ntf_dunn(order=3, osr=64, H_inf=1.5):
    """
    Alias of :func:`ntf_dunn`

    .. deprecated:: 0.11.0
       Function has been moved to the :mod:`pydsm.NTFdesign` module with
       name :func:`pydsm.NTFdesign.ntf_dunn`.
    """
    warn("Function superseded by ntf_dunn in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_dunn(order, osr, H_inf)


def synthesize_ntf_from_audio_weighting(
        order, osr,
        audio_weighting=audio_weightings.f_weighting,
        audio_band=22.05E3,
        max_attn=120,
        H_inf=1.5,
        normalize="auto", **options):
    """
    Alias of :func:`ntf_fir_audio_weighting`

    .. deprecated:: 0.11.0
        Function has been moved to the :mod:`pydsm.NTFdesign` module with
        name :func:`pydsm.NTFdesign.ntf_fir_audio_weighting`.
    """
    warn("Function superseded by ntf_fir_audio_weighting in "
         "NTFdesign module", PyDsmDeprecationWarning)
    return ntf_fir_audio_weighting(order, osr, audio_weighting, audio_band,
                                   max_attn, H_inf, normalize, **options)



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from ... import cvxpy_tdr


def ntf_fir_from_digested(Qs, A, C, H_inf, **opts):
    """
    Synthesize FIR NTF from predigested specification

    Version for the cvxpy_tdr modeler.
    """
    quiet = not opts['show_progress']
    order = np.size(Qs, 0)-1
    br = cvxpy_tdr.variable(order, 1, name='br')
    b = cvxpy_tdr.vstack((1, br))
    X = cvxpy_tdr.variable(order, order, structure='symmetric', name='X')
    Qs = cvxpy_tdr.matrix(Qs)
    target = cvxpy_tdr.norm2(Qs*b)
    A = cvxpy_tdr.matrix(A)
    B = cvxpy_tdr.vstack((cvxpy_tdr.zeros((order-1, 1)), 1.))
    C = cvxpy_tdr.matrix(C)
    C = C+(cvxpy_tdr.matrix(np.eye(order, order)[:, ::-1])*br).T
    D = cvxpy_tdr.matrix(1.)
    M1 = A.T*X
    M2 = M1*B
    M = cvxpy_tdr.vstack((
        cvxpy_tdr.hstack((M1*A-X, M2, C.T)),
        cvxpy_tdr.hstack((M2.T, B.T*X*B-H_inf**2, D)),
        cvxpy_tdr.hstack((C, D, -1))
        ))
    constraint1 = cvxpy_tdr.belongs(-M, cvxpy_tdr.semidefinite_cone)
    constraint2 = cvxpy_tdr.belongs(X, cvxpy_tdr.semidefinite_cone)
    p = cvxpy_tdr.program(cvxpy_tdr.minimize(target),
                             [constraint1, constraint2])
    p.options.update(opts["cvxpy_tdr_opts"])
    p.solve(quiet)
    return np.hstack((1, np.asarray(br.value.T)[0]))



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from .expand import expand
from ..defs import *
import scipy.sparse as sp
from cvxopt import solvers
from ..scalars import cvxpy_obj
from ..constraints import cvxpy_list

# Function: call_solver
def call_solver(p,quiet):
    """
    Calls solver.

    :param p: Convex cvxpy_program.
              Assumed to be expanded.
    :param quiet: Boolean.
    """

    # Set printing format for cvxopt sparse matrices
    opt.spmatrix_str = opt.printing.spmatrix_str_triplet

    # Expand objects defined via partial minimization
    constr_list = cvxpy_list(pm_expand(p.constraints))

    # Get variables
    variables = constr_list.variables
    variables.sort()

    # Count variables
    n = len(variables)

    # Create variable-index map
    var_to_index = {}
    for i in range(0,n,1):
        var_to_index[variables[i]] = i

    # Construct objective vector
    c = construct_c(p.objective,var_to_index,n,p.action)

    # Construct Ax == b
    A,b = construct_Ab(constr_list._get_eq(),var_to_index,n)

    # Construct  Gx <= h
    G,h,dim_l,dim_q,dim_s = construct_Gh(constr_list._get_ineq_in(),
                                         var_to_index,n)

    # Construct F
    F = construct_F(constr_list._get_ineq_in(),var_to_index,n)

    # Call cvxopt
    solvers.options['maxiters'] = p.options['maxiters']
    solvers.options['abstol'] = p.options['abstol']
    solvers.options['reltol'] = p.options['reltol']
    solvers.options['feastol'] = p.options['feastol']
    solvers.options['show_progress'] = not quiet
    dims = {'l':dim_l, 'q':dim_q, 's':dim_s}
    if F is None:
        r =  solvers.conelp(c,G,h,dims,A,b)
    else:
        r =  solvers.cpl(c,F,G,h,dims,A,b)

    # Store numerical values
    if r['status'] != PRIMAL_INFEASIBLE:
        for v in variables:
            v.value =  r['x'][var_to_index[v]]

    # Return result
    return r

# Function: construct_c
def construct_c(objective,mapping,n,action):
    """
    Creates the vector c for the objective function c.T*x

    :param objective: cvxpy_scalar_var or cvxpy_obj.
    :param mapping: Dictionary.
    :param n: Size for vector c.
    :param action: MINIMIZE OR MAXIMIZE
    """

    # Check objective
    if (objective.type != CONSTANT and
        objective.type != VARIABLE):
        raise TypeError('Bad objective: Cannot construct c')

    # Construct c vector
    c = opt.matrix(0.0,(n,1))
    obj_vars = objective.variables
    if not len(obj_vars):
        return c
    elif len(obj_vars) == 1:
        obj_var = obj_vars[0]
        i = mapping[obj_var]
        if action == MINIMIZE:
            c[i] = 1.
        else:
            c[i] = -1.
    else:
        raise TypeError('Bad objective: Cannot construct c')

    # Return vector
    return c

# Function: construct_Ab
def construct_Ab(constr_list,mapping,n):
    """
    Constructs matrix A and vector b from a list
    of equality constraints.

    :param constr_list: List of equality constraints.
    :param mapping: Dictionary.
    :param n: Number of variables.
    """

    # Sort constraints
    def cmp(x,y):
        lx = len(x.variables)
        if ((type(x.left)  is cvxpy_obj and x.left.value != 0) or
            (type(x.right) is cvxpy_obj and x.right.value != 0)):
             lx += 1
        ly = len(y.variables)
        if ((type(y.left)  is cvxpy_obj and y.left.value != 0) or
            (type(y.right) is cvxpy_obj and y.right.value != 0)):
            ly += 1
        return 1 if lx >= ly else -1

    # Sort constraints keys
    def cmp_keys(x):
        lx = len(x.variables)
        if ((type(x.left)  is cvxpy_obj and x.left.value != 0) or
            (type(x.right) is cvxpy_obj and x.right.value != 0)):
            lx += 1
        return lx

    constr_list.sort(key=cmp_keys)

    # Construct matrices
    Ab = None
    for constr in constr_list:

        # Get elements
        left = constr.left
        right = constr.right

        # New row
        t = sp.lil_matrix((1,n+1))

        # Deal with right element
        if right.type == CONSTANT:
            t[0,n] += right.value*1.
        elif right.type == VARIABLE:
            t[0,mapping[right]] += -1.
        else:
            raise TypeError('Bad equality: Cannot construct A,b')

        # Left is a variable
        if left.type == VARIABLE:
            t[0,mapping[left]] += 1.

        # Left is a constant
        elif left.type == CONSTANT:
             t[0,n] += -left.value*1.

        # Left is an operation tree
        elif left.type == TREE and left.item.type == OPERATOR:

            # Addition
            if left.item.name == SUMMATION:
                for arg in left.children:
                    if (arg.type == TREE and
                        arg.item.type == OPERATOR and
                        arg.item.name == MULTIPLICATION):
                        ch1 = arg.children[1]
                        ch0 = arg.children[0]
                        t[0,mapping[ch1]] += ch0.value*1.
                    elif arg.type == VARIABLE:
                        t[0,mapping[arg]] += 1.
                    elif arg.type == CONSTANT:
                        t[0,n] += -arg.value*1.
                    else:
                        raise TypeError('Bad equality: Cannot construct A,b')

            # Multiplication
            elif left.item.name == MULTIPLICATION:
                op1 = left.children[0]
                op2 = left.children[1]
                t[0,mapping[op2]] += 1.*op1.value

            # Error
            else:
                raise TypeError('Bad equality: Cannot construct A,b')

        # Error
        else:
            raise TypeError('Bad equality: Cannot construct A,b')

        # Compute norm
        t = t.tocsr()
        t_norm = np.sqrt((t*t.T)[0,0])

        # Process
        if t_norm < EPSILON:
            continue
        if Ab is None:
            Ab = t
        else:
            t_c = t.nonzero()[1]
            Ab_c = Ab.nonzero()[1]
            c_overlap = True
            for x in t_c:
                if x not in Ab_c:
                    c_overlap = False
                    break
            valid_row = True
            if c_overlap:
                tn = t/t_norm
                for i in range(Ab.shape[0]-1,-1,-1):
                    s = Ab[i,:]
                    s_c = s.nonzero()[1]
                    if len(s_c) < len(t_c):
                        break
                    if set(s_c) == set(t_c):
                        e = s-(tn*s.T)*tn
                        if np.sqrt((e*e.T)[0,0]) < EPSILON:
                            valid_row = False
                            break
            if valid_row:
                Ab = sp.vstack((Ab,t)).tocsr()

    # Return matrices
    if Ab is None:
        Ab = opt.spmatrix(0.0,[],[],(0,n+1))
    else:
        nz = Ab.nonzero()
        Ab = opt.spmatrix(Ab.data.tolist(),nz[0].tolist(),
                          nz[1].tolist(),Ab.shape)
    return Ab[:,0:n],opt.matrix(Ab[:,n])

# Function: construct_Gh
def construct_Gh(constr_list,mapping,n):
    """
    Creates the matrix G and vector h from a list
    of inequality and membership constraints.

    :param constr_list: List of inequality and membership constraints.
    :param mapping: Dictionary.
    :param n: Number of varaibles.
    """

    # Check constraints
    for c in constr_list:
        if c.left.type == ARRAY:
            if c.right.type != SET:
                raise TypeError('Bad constraint: Cannot construct G,h')
        else:
            if ((c.left.type != VARIABLE and c.left.type != CONSTANT) or
                (c.right.type != VARIABLE and c.right.type != CONSTANT)):
                raise TypeError('Bad constraint: Cannot construct G,h')

    # Initialize G,h and dimensions
    dim_l = 0
    dim_q = []
    dim_s = []
    G = opt.spmatrix(0.0,[],[],(0,n))
    h = opt.matrix(0.0,(0,1))

    # Nonnegative Orthant
    for c in constr_list:

        # Left is constant or variable
        if c.left.type == CONSTANT or c.left.type == VARIABLE:

            # Get constraint elements
            t = c.type
            ob1 = c.left
            ob2 = c.right

            # New row
            rowG = opt.spmatrix(0.0,[],[],(1,n))
            rowh = opt.matrix(0.0,(1,1))
            if ob1.type == VARIABLE:
                rowG[0,mapping[ob1]]+=1. if t==LESS_EQUALS else -1.
            else:
                rowh[0,0]+=-ob1.value*1. if t==LESS_EQUALS else ob1.value*1.
            if ob2.type == VARIABLE:
                rowG[0,mapping[ob2]]+=1. if t==GREATER_EQUALS else -1.
            else:
                rowh[0,0]+=-ob2.value*1. if t==GREATER_EQUALS else ob2.value*1.

            # Append to G,h
            G = opt.sparse([G,rowG])
            h = opt.matrix([h,rowh])

            # Increment size of cone
            dim_l += 1

    # Second order cone
    for c in constr_list:

        # Set membership that expands to SOC
        if (c.right.type == SET and
            c.right.expansion_type == SOC):

            # Get G,h, section
            el = c.left
            set_atom = c.right
            newG,newh,r = set_atom._construct(el,mapping,n)

            # Attach to G,h
            G = opt.sparse([G,newG])
            h = opt.matrix([h,newh])

            # Attach size of cone
            dim_q = dim_q + [r]

    # Semidefinite cone
    for c in constr_list:

        # Set membership that expands to LMI
        if (c.right.type == SET and
            c.right.expansion_type == SDC):

            # Get G,h, section
            el = c.left
            set_atom = c.right
            newG,newh,t = set_atom._construct(el,mapping,n)

            # Attach to G,h
            G = opt.sparse([G,newG])
            h = opt.matrix([h,newh])

            # Attach size of cone
            dim_s = dim_s + [t]

    # Return
    return G,h,dim_l,dim_q,dim_s

# Function
def construct_F(constr_list,mapping,n):
    """
    Constructs function F needed by cvxopt
    to solve nonlinear programs.

    :param constr_list: List of inequality and set
                        membership constraints.
    :param mapping: Dictionary.
    :param n: Numbe of variables.
    """

    # Check constraints
    for c in constr_list:
        if c.left.type == ARRAY:
            if c.right.type != SET:
                raise TypeError('Bad constraint: Cannot construct F')
        else:
            if ((c.left.type != VARIABLE and c.left.type != CONSTANT) or
                (c.right.type != VARIABLE and c.right.type != CONSTANT)):
                raise TypeError('Bad constraint: Cannot construct F')

    # Lists
    fs = []
    grads = []
    hess = []
    inds = []

    for c in constr_list:

        # Set that has a (f,gradf,hessf) implementation
        if (c.right.type == SET and
            c.right.expansion_type == DIF):

            # Get objects
            el = c.left
            set_atom = c.right

            # Get functions
            n_f,n_grad,n_hess,n_ind = set_atom._construct(el,mapping,n)

            # Append
            fs = fs + [n_f]
            grads = grads + [n_grad]
            hess = hess + [n_hess]
            inds = inds + [n_ind]

    # Check if no functions
    if not len(fs):
        return None

    # Construct F
    def F(x=None,z=None):

        # Case 1
        if x is None and z is None:
            x0 = opt.matrix(1., (n,1))
            return len(fs),x0

        # Case 2
        elif x is not None and z is None:
            if all(list(map(lambda y: y(x),inds))):
                f = opt.matrix(0.0,(len(fs),1))
                for i in range(0,len(fs),1):
                    f[i] = fs[i](x)
                Df = opt.spmatrix(0.0,[],[],(0,n))
                for i in range(0,len(grads),1):
                    Df = opt.sparse([Df,grads[i](x).T])
                return f,Df
            else:
                return None,None

        # Case 3
        else:
            f = opt.matrix(0.0,(len(fs),1))
            for i in range(0,len(fs),1):
                f[i] = fs[i](x)
            Df = opt.spmatrix(0.0,[],[],(0,n))
            for i in range(0,len(grads),1):
                Df = opt.sparse([Df,grads[i](x).T])
            H = opt.spmatrix(0.0,[],[],(n,n))
            for i in range(0,len(hess),1):
                H = H + z[i]*hess[i](x)
            return f,Df,H

    # Return F
    return F

# Function: Partial minimization expansion
def pm_expand(constr_list):
    """
    Expands functions that are implemented
    using partial minimization descriptions.

    :param constr_list: List of constraints.
    """

    new_list = []
    for c in constr_list:
        if (c.left.type == TREE and
            c.left.item.type == FUNCTION):
            new_constr = expand(c.left.item._pm_expand(c))
            new_list += pm_expand(new_constr)
        elif (c.right.type == SET and
              c.right.expansion_type == PM):
            new_constr = expand(c.right._pm_expand(c))
            new_list += pm_expand(new_constr)
        else:
            new_list += [c]

    # Return new list
    return new_list



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Legacy NTF synthesis methods (:mod:`pydsm.NTFdesign.legacy`)
============================================================

This modules provides code for the synthesis of the modulator NTF, based
on methods that are now superseded but that are worth keeping around for
reproducing published results.

.. currentmodule:: pydsm.NTFdesign.legacy


Functions
---------

.. autosummary::
   :toctree: generated/

   q0_from_filter_ir                -- Legacy `q0_weighting`
   quantization_noise_gain_by_conv  -- Legacy `quantization_noise_gain`
"""

from ._quantization_noise_gain import *
from ._q0_from_ir import *

__all__ = (_quantization_noise_gain.__all__ + _q0_from_ir.__all__
           + ['test'])

from ..._pytesttester import PytestTester
test = PytestTester(__name__)
del PytestTester



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from scipy import signal
from pydsm.ir import impulse_response
from pydsm.NTFdesign import (ntf_fir_weighting, quantization_noise_gain,
                             mult_weightings)
from pydsm.NTFdesign.weighting import ntf_fir_from_q0
from pydsm.NTFdesign.legacy import (quantization_noise_gain_by_conv,
                                    q0_from_filter_ir)
import pytest

__all__ = ["TestNTF_Filter"]


class TestNTF_Filter:

    @classmethod
    def setup_class(cls):
        z_e = np.asarray([0.98979462+0.12667657j,
                          0.98979462-0.12667657j,
                          0.72084151+0.0j,
                          0.35347507+0.64857031j,
                          0.35347507-0.64857031j,
                          -0.02875404+0.70480695j,
                          -0.02875404-0.70480695j,
                          -0.36294495+0.58281858j,
                          -0.36294495-0.58281858j,
                          -0.67350105+0.0j,
                          -0.59201143+0.32765994j,
                          -0.59201143-0.32765994j])
        cls.z_e = np.sort(z_e)
        cls.order = z_e.size
        # Generate filter.
        # 8th order bandpass filter
        # Freq. passed to butterworth is normalized between 0 and 1
        # where 1 is the Nyquist frequency
        fsig = 1000.
        B = 400.
        OSR = 64
        fphi = B*OSR*2
        w0 = 2*fsig/fphi
        B0 = 2*B/fphi
        w1 = (np.sqrt(B0**2+4*w0**2)-B0)/2
        w2 = (np.sqrt(B0**2+4*w0**2)+B0)/2
        cls.hz = signal.butter(4, [w1, w2], 'bandpass', output='zpk')

    def test_ntf_butt_bp8_vs_legacy(self):
        modeler = 'cvxpy'
        try:
            import cvxpy
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        # Compute q0 in two ways
        ir = impulse_response(self.hz, db=80)
        ntf1 = ntf_fir_weighting(self.order, self.hz, show_progress=False,
                                 modeler=modeler)
        q0 = q0_from_filter_ir(self.order, ir)
        ntf2 = ntf_fir_from_q0(q0, show_progress=False,
                               modeler=modeler)
        mf1 = quantization_noise_gain(ntf1, self.hz)
        mf2 = quantization_noise_gain(ntf2, self.hz)
        np.testing.assert_allclose(mf2, mf1, rtol=1e-7)
        mf3 = quantization_noise_gain_by_conv(ntf1, self.hz)
        np.testing.assert_allclose(mf3, mf1, rtol=5e-6)

    def test_ntf_butt_bp8_cvxpy_tdr(self):
        try:
            from pydsm import cvxpy_tdr     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy_old' not installed")
        ntf1 = ntf_fir_weighting(self.order, self.hz, modeler='cvxpy_old',
                                 show_progress=False)
        np.testing.assert_allclose(np.sort(ntf1[0]), self.z_e, rtol=1e-5)

    def test_ntf_butt_bp8_cvxpy_cvxopt(self):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        ntf1 = ntf_fir_weighting(self.order, self.hz, modeler='cvxpy',
                                 show_progress=False)
        np.testing.assert_allclose(np.sort(ntf1[0]), self.z_e, rtol=1e-7)

    def test_ntf_butt_bp8_cvxpy_scs(self):
        try:
            import cvxpy     # analysis:ignore
        except:
            pytest.skip("Modeler 'cvxpy' not installed")
        ntf1 = ntf_fir_weighting(self.order, self.hz, modeler='cvxpy',
                                 cvxpy_opts={'solver': 'scs'},
                                 show_progress=False,
                                 scs_opts={"eps": 1E-10,
                                           "max_iters": 100000,
                                           "acceleration_lookback": 10})
        np.testing.assert_allclose(np.sort(ntf1[0]), self.z_e, rtol=1e-4)

    def test_ntf_butt_bp8_picos(self):
        try:
            import picos     # analysis:ignore
        except:
            pytest.skip("Modeler 'picos' not installed")
        ntf1 = ntf_fir_weighting(self.order, self.hz, modeler='picos',
                                 show_progress=False)
        np.testing.assert_allclose(np.sort(ntf1[0]), self.z_e, rtol=1e-7)


class Test_MultWeightings:

    def setUp(self):
        pass

    def test_mult_weightings(self):
        f = mult_weightings(([], [], 1), ([], [], 2))
        np.testing.assert_equal(f(0), 4)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Correlation utilities (:mod:`pydsm.correlations`)
=================================================

Functions to compute the auto- and cross- correlation between two vectors.

So far only unnormalized (raw) correlations are supported.

.. currentmodule:: pydsm.correlations

Functions
---------

.. autosummary::
   :toctree: generated/

    raw_acorr  -- raw autocorrelation of a vector
    raw_xcorr  -- raw crosscorrelation of a vector
"""

from __future__ import division, print_function

import numpy as np

import sys
if sys.version_info < (3,):
    range = xrange

__all__ = ["raw_acorr", "raw_xcorr"]


def raw_acorr(x, N):
    """
    Computes the raw autocorrelation of a vector up to lag N.

    Parameters
    ----------
    x : array_like
        1-D sequence to compute the auto-correlation upon
    N : int
        the maximum (positive) lag of the raw auto-correlation to return.

    Returns
    -------
    q : ndarray
        the raw (unnormalized) autocorrelation vector.
        Assuming that :math:`m` is the length of x,
        :math:`q(k) = \\sum_{n=k}^{m-1} x(n) x(n-k)
        \\text{ for } k = 0 \\dots N`

    Notes
    -----
    The routine does not make any check on the length of x and N. It
    is responsibility of the user to assure that len(x)>=N. In some cases
    (but only in some cases), zero padding is practiced.
    """
    m = len(x)
    q = np.asarray([np.dot(x[k:m], x[0:m-k]) for k in range(N+1)])
    return q


def raw_xcorr(x, y, N):
    """
    Computes the raw crosscorrelation between two vectors up to lag N.

    Parameters
    ----------
    x : array_like
        first 1-D vector
    y : array_like
        second 1-D vector
    N : int
        the maximum (positive) lag of the raw cross-correlation to return.

    Returns
    -------
    q : ndarray
        the raw (unnormalized) crosscorrelation vector.
        Assuming that :math:`m_x` and :math:`m_y` are the lengths of x and y,
        :math:`q(k) = \\sum_{n=k}^{\\min(m_x-1,m_y+k-1)}
        x(n) y(n-k) \\text{ for } k = 0 \\dots N`

    Notes
    -----
    the routine does not make any check on the lengths of x, y and N. It
    is responsibility of the user to assure that N<=len(y). In some cases
    (but only in some cases), zero padding is assumed.
    """
    mx = len(x)
    my = len(y)
    q = np.asarray([np.dot(y[k:min(my, mx+k)],
                           x[0:min(my-k, mx)]) for k in range(N+1)])
    return q



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
DELSIG-like synthesizeChebyshevNTF
==================================
"""

import numpy as np
from scipy.signal import cheby2
from ._ds import ds_f1f2
from ..relab import cplxpair

import sys
if sys.version_info < (3,):
    range = xrange

__all__ = ["synthesizeChebyshevNTF"]


def synthesizeChebyshevNTF(order=3, osr=64, opt=0, H_inf=1.5, f0=0.0):
    """
    Synthesize the NTF for a DS modulator by a Chebyshev function.

    This function is an alternative to synthesizeNTF which assumes
    that the magnitude of the denominator of the NTF is approximately constant
    in the passband. When the OSR or H_inf are low, this assumption breaks
    down and synthesizeNTF yields a non-optimal NTF. The synthesizeChebyshevNTF
    creates non-optimal NTFs, but fares better than synthesizeNTF in the
    aforementioned circumstances. The NTF is obtained from a type-2 highpass
    Chebyshev function.

    Parameters
    ----------
    order : int, optional
        the order of the modulator, defaults to 3
    osr : float, optional
        the oversamping ratio (based on the actual signal bandwidth)
    opt : int, optional
        ignored, present for signature compatibility with synthesizeNTF
    H_inf : real, optional
        max allowed peak value of the NTF. Defaults to 1.5
    f0 : real, optional
        center frequency, defaults to 0.
        1 corresponds to the sampling frequency, so that 0.5 is the
        maximum value. Value 0 specifies an LP modulator.

    Returns
    -------
    ntf : tuple
        noise transfer function in zpk form.

    Raises
    ------
    ValueError
        'Order must be even for a bandpass modulator.' if the order is
        incompatible with the modulator type.

    Notes
    -----
    Parameter opt is ignored.
    """
    if f0 != 0 and order % 2 != 0:
        raise ValueError('Order must be even for a bandpass modulator.')
    else:
        f1, f2 = ds_f1f2(osr, f0)
        f1f2 = (f1, f2)

    # Iteratively solve for the attenuation spec (x) which yields the
    # desired H_inf
    x_min = 0.
    x_max = 300.
    dx_max = 10.
    ftol = 1e-6
    xtol = 1e-6
    # Initial guess for x
    x = 60.
    itn_limit = 10
    converged = False
    f_p = 0.

    # Design is based on a Chebyshev type II filter, designed with the cheby2
    # function. Recall that a Chebyshev type II filter is monothonic in the
    # pass-band and equiripple in the stop-band.
    # In cheby2 parameters are:
    #   N : order
    #   Rs : stopband attenuation
    #        this is measured from the peak value in the pass-band
    #        to the peak(s) of the stop-band, in dB
    #   Wn : stopband edge frequency (normalized so that 1 is fs/2 for DT
    #        filters). Must be a couple of frequencies for stop band filters
    #   btype : filter type (e.g. 'low', 'high', 'stop')
    #   analog : non zero for a CT filter (defaults to DT)
    #   output : 'ba' or 'zpk'
    # The filter that is returned has peak gain at 1 in the pass-band

    for itn in range(itn_limit):
        if f0 == 0:
            z, p, k = cheby2(order, x, 1./osr, 'high', output='zpk')
        else:
            z, p, k = cheby2(order/2, x, 2*f1f2, 'stop',
                             output='zpk')
        f = 1./k - H_inf
        # print (x, f)
        if f > 0:
            # x is too big
            x_max = x
        else:
            # x is too small
            x_min = x
        if itn == 0:
            # First iteration
            dx = -dx_max * np.sign(f)
        else:
            # Normal iteration
            df = f - f_p
            if np.abs(df) < ftol:
                converged = True
                break
            dx = -f * dx/df
        if converged:
            break
        x_p = x
        f_p = f
        x = np.max((x_min, np.min((x+dx, x_max))))
        dx = x - x_p
        if np.abs(dx) < xtol:
            break

    z = cplxpair(z)
    p = cplxpair(p)
    return (z, p, 1)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Slow simulator for a generic delta sigma modulator
==================================================
"""

import numpy as np
from scipy.signal import zpk2ss
from scipy import linalg
from warnings import warn
from ..exceptions import PyDsmSlowPathWarning

import sys
if sys.version_info < (3,):
    range = xrange


__all__ = ["ds_quantize"]


def simulateDSM(u, arg2, nlev=2, x0=0,
                store_xn=False, store_xmax=False, store_y=False):

    warn('Running the slow version of simulateDSM.',
         PyDsmSlowPathWarning)

    # Make sure that nlev is an array
    nlev = np.asarray(nlev).reshape(1)

    # Make sure that input is a matrix
    u = np.asarray(u)
    if u.ndim == 1:
        u = u.reshape(1, -1)

    nu = u.shape[0]
    nq = np.size(nlev)

    if type(arg2) == tuple and len(arg2) == 3:
        # Assume ntf in zpk form
        (ntf_z, ntf_p, ntf_k) = arg2
        form = 2
        order = len(ntf_z)
    else:
        # Assume ABCD form
        ABCD = np.asarray(arg2)
        if ABCD.shape[1] == nu+ABCD.shape[0]:
            # ABCD dimensions OK
            form = 1
            order = ABCD.shape[0]-nq
        else:
            raise ValueError('Incorrect modulator specification')

    # Assure that the state is a column vector
    if np.isscalar(x0) and x0 == 0:
        x0 = np.zeros((order, 1))
    else:
        x0 = np.array(x0, dtype=float).reshape(-1, 1)

    if form == 1:
        A = ABCD[0:order, 0:order]
        B = ABCD[0:order, order:order+nu+nq]
        C = ABCD[order:order+nq, 0:order]
        D1 = ABCD[order:order+nq, order:order+nu]
    else:
        # Seek a realization of -1/H
        A, B2, C, D2 = zpk2ss(ntf_p, ntf_z, -1)
        # Transform the realization so that C = [1 0 0 ...]
        Sinv = (linalg.orth(np.hstack((np.transpose(C), np.eye(order)))) /
                np.linalg.norm(C))
        S = linalg.inv(Sinv)
        C = np.dot(C, Sinv)
        if C[0, 0] < 0:
            S = -S
            Sinv = -Sinv
        A = np.dot(np.dot(S, A), Sinv)
        B2 = np.dot(S, B2)
        C = np.hstack(([[1]], np.zeros((1, order-1))))
        # C=C*Sinv;
        # D2 = 0;
        # !!!! Assume stf=1
        B1 = -B2
        D1 = 1
        B = np.hstack((B1, B2))

    N = u.shape[1]
    v = np.empty((nq, N))
    if store_y:
        # Need to store the quantizer input
        y = np.empty((nq, N))
    else:
        y = np.empty((0, 0))
    if store_xn:
        # Need to store the state information
        xn = np.empty((order, N))
    if store_xmax:
        # Need to keep track of the state maxima
        xmax = np.abs(x0)
    else:
        xmax = np.empty(0)

    for i in range(N):
        # I guess the coefficients in A, B, C, D should be real...
        y0 = np.real(np.dot(C, x0) + np.dot(D1, u[:, i]))
        if store_y:
            y[:, i] = y0
        v[:, i] = ds_quantize(y0, nlev)
        x0 = np.dot(A, x0) + np.dot(B, np.vstack((u[:, i], v[:, i])))
        if store_xn:
            # Save the next state
            xn[:, i] = x0
        if store_xmax:
            # Keep track of the state maxima
            xmax = np.max((np.abs(x0), xmax), 0)
    if not store_xn:
        xn = x0
    return v.squeeze(), xn.squeeze(), xmax, y.squeeze()


def ds_quantize(y, n):
    """Quantize a signal according to a given number of levels.

    Parameters
    ----------
    y : real or array of reals
        signal to be quantized (1 sample!). A column vector with more than
        1 row if there are multiple quantizers.
    n : int or vector of ints
        number of quantization levels. Can be a vector to specify multiple
        quantizers, in this case, y must have as many rows as the entries in
        n

    Returns
    -------
    z : real or ndarray
        quantized signal (1 sample!). A column vector with
        more than 1 row if there are multiple quantizers.

    Notes
    -----
    y is quantized to:

    * an odd integer in [-n+1, n-1], if n is even, or
    * an even integer in [-n, n], if n is odd.

    This definition gives the same step height for both mid-riser and
    mid-tread quantizers.
    """
    v = np.empty_like(y)
    for qi in range(np.size(n)):
        if np.remainder(n[qi], 2) == 0:
            v[qi] = 2*np.floor(0.5*y[qi])+1
        else:
            v[qi] = 2*np.floor(0.5*(y[qi]+1))
        L = n[qi]-1
        v[qi, 0] = np.max((np.min((v[qi, 0], L)), -L))
    return v



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

"""
Entry point for DELSIG-like Delta-Sigma modulator simulator
===========================================================
"""

from ._simulateDSM_scipy import simulateDSM as _simulateDSM_scipy
try:
    from ._simulateDSM_cblas import simulateDSM as _simulateDSM_cblas
    HAS_CBLAS = True
except:
    HAS_CBLAS = False
from ._simulateDSM_scipy_blas import simulateDSM as _simulateDSM_scipy_blas
from ..utilities import digested_options

__all__ = ["simulateDSM"]


def simulateDSM(u, arg2, nlev=2, x0=0,
                store_xn=False, store_xmax=False, store_y=False,
                **options):
    """
    Computes the output of a general delta-sigma modulator.

    Parameters
    ----------
    u : array_like or matrix_like
        modulator input. Multiple inputs are allowed. In this case,
        u is a matrix with as many rows as the desired inputs
    arg2 : tuple
        modulator structure in ABDC matrix form or modulator NTF
        as zpk tuple. In the latter case, the modulator STF is assumed
        to be unitary.
    nlev : int or array of ints, optional
        number of levels in quantizer. Multiple quantizers can be
        specified by making nlev a vector. Defaults to 2.
    x0 : array_like of reals or 0
        modulator intitial state vector. Assigning it to 0 is a shorthand
        for an appropriate length zero vector. Defaults to 0.
    store_xn : bool, optional
        switch controlling the storage of state evolution.
        See description of return values. Defaults to False.
    store_xmax : bool, optional
        switch controlling the storage of maxima in state variables.
        See description of return values. Defaults to False.
    store_y : bool, optional
        switch controlling the storage of quantizer input values.
        See description of return values. Defaults to False.

    Returns
    -------
    v : ndarray
        samples at the output of the modulator, one per input sample.
        If there are multiple quantizers, then v is a matrix, with as many
        columns as the number of samples and as many rows as the number of
        quantizers.
    xn : ndarray
        internal state of the modulator. If store_xn is set to True, then
        it includes a state snapshot per input sample. In this case, xn
        is a matrix, with as many columns as the number of samples and as
        many rows as the number of state variables. If store_xn is False,
        then xn is a vector containing a snapshot of the last state.
    xmax : ndarray
        maximum absolute value reached by the state variables.
        If store_xmax is set to True, then xmax is a vector with as many
        entries as the number of state variables. Otherwise it is null.
    y : ndarray
        samples at the quantizer input(s), one per input sample.
        If store_y is set to True, then y records the quantizer(s)
        input(s). If there are multiple quantizers, then y is a matrix,
        with as many columns as the number of samples and as many rows as
        the number of quantizers. If store_y is False, then y is null.

    Other Parameters
    ----------------
    backend : string
        Use: 'auto' for automatic selection; 'scipy' for pure python
        simulator; 'cblas' for simulator using platform cblas library;
        'scipy_blas' for simulator using scipy provided blas. Defaults can
        be set by changing the function ``default_options`` attribute.

    Raises
    ------
    ValueError
        'Incorrect modulator specification', if the modulator specification
        is inconsistent.

        'Invalid argument: nlev must be convertible into a 1D int array',
        if the quantizer specification is incorrect.

        'Invalid argument: u must be convertible into a 2D float array',
        if the input specification is incorrect.

        'Incorrect initial condition specification' if the initial condition
        specification for the modulator filters is incorrect.

    RuntimeError
        'Unsupported simulator backend xxx' if an unsupported backend is
        required

    Warns
    -----
    PyDsmSlowPathWarning
        'Running the slow version of simulateDSM', if the simulator being
        used is the slow one, coded in pure Python.

    Notes
    -----
    The quantizer is ideal, producing integer outputs centered about zero.
    Quantizers with an even number of levels are of the mid-rise type and
    produce outputs which are odd integers. Quantizers with an odd number of
    levels are of the mid-tread type and produce outputs which are even
    integers.

    The modulator structure being simulated is a block diagonal one, as
    returned by the zpk2ss function.

    Setting store_xn, store_xmax and store_y to False speeds up the operation.

    There are actually two simulators, sharing this function as a front end.
    One of them is coded in pure python and quite slow. The other one is
    coded in C (actually in Cython), and directly accesses low level cblas
    functions. The codebase to be used is controlled by the module switch
    `use_fast_simulator`.
    """
    # Manage options
    opts = digested_options(options, simulateDSM.default_options,
                            ['backend'])
    backend = opts["backend"]
    if backend == 'auto':
        simulator = _simulateDSM_scipy_blas
    elif backend == 'scipy':
        simulator = _simulateDSM_scipy
    elif backend == 'scipy_blas':
        simulator = _simulateDSM_scipy_blas
    elif backend == 'cblas' and HAS_CBLAS:
        simulator = _simulateDSM_cblas
    else:
        raise RuntimeError('Unsupported simulator backend %s' % backend)
    return simulator(u, arg2, nlev, x0, store_xn, store_xmax, store_y)

simulateDSM.default_options = {'backend': 'auto'}



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..sets import *
from ..utils import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Nuclear norm
def nuclear_norm(X):
    r"""
    | :math:`\mbox{nuclear\_norm} :
      \mathbb{R}^{m \\times n} \\to \mathbb{R},
      \ \mbox{nuclear\_norm}(X) = \mbox{Tr} \sqrt{X^TX}`.
    | Convex.

    :param X: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check input
    if (np.isscalar(X) or
        type(X).__name__ in SCALAR_OBJS):
        X = vstack([X])
    elif (type(X) is not cvxpy_matrix and
          type(X).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Construct program
    (m,n) = X.shape
    Y = variable(m,m,'symmetric')
    W = variable(n,n,'symmetric')
    Z = variable(m,n)
    p = program(minimize((1./2.)*(trace(Y)+trace(W))),
                [belongs(vstack((hstack((Y,Z)),
                                 hstack((Z.T,W)))),
                         semidefinite_cone)],
                [Z],
                name='nuclear_norm')
    return p(X)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import sum
from ..arrays import *
from ..scalars import *
from ..constraints import *
from ..interface import equals
from ..interface import belongs
from ..interface import less_equals
from ..interface import greater_equals

# Function: re_eval
def re_eval(arg,replace_map):
    """
    Re-evaluates argument.

    :param replace_map: Dictionary.
    """

    # Number
    if np.isscalar(arg):
        return arg

    # Constant object
    elif type(arg) is cvxpy_obj:
        return arg.value

    # Scalar variable
    elif type(arg) is cvxpy_scalar_var:
        if arg in replace_map.keys():
            return re_eval(replace_map[arg],replace_map)
        else:
            return arg

    # Scalar param
    elif type(arg) is cvxpy_scalar_param:
        if arg in replace_map.keys():
            return re_eval(replace_map[arg],replace_map)
        else:
            return arg

    # Summation
    elif (type(arg) is cvxpy_tree and
          arg.item.type == OPERATOR and
          arg.item.name == SUMMATION):
        new_children = list(map(lambda x:re_eval(x,replace_map),arg.children))
        return sum(new_children)

    # Multiplication
    elif (type(arg) is cvxpy_tree and
          arg.item.type == OPERATOR and
          arg.item.name == MULTIPLICATION):
        child1 = re_eval(arg.children[0],replace_map)
        child2 = re_eval(arg.children[1],replace_map)
        return child1*child2

    # Function
    elif (type(arg) is cvxpy_tree and
          arg.item.type == FUNCTION):
        new_children = list(map(lambda x:re_eval(x,replace_map),arg.children))
        return arg.item(new_children)

    # Constraint
    elif type(arg) is cvxpy_constr:

        # Not set membership
        if arg.type != BELONGS:
            left = re_eval(arg.left ,replace_map)
            right= re_eval(arg.right,replace_map)
            if arg.type == EQUALS:
                return equals(left,right)
            elif arg.type == LESS_EQUALS:
                return less_equals(left,right)
            elif arg.type == GREATER_EQUALS:
                return greater_equals(left,right)
            else:
                raise TypeError('Invalid constraint')

        # Set membership
        else:
            left = re_eval(arg.left,replace_map)
            return belongs(left,arg.right)

    # Array
    elif (type(arg) is cvxpy_array or
          type(arg) is cvxpy_var or
          type(arg) is cvxpy_param):
        (m,n) = arg.shape
        new_ar = cvxpy_array(m,n)
        for i in range(0,m,1):
            for j in range(0,n,1):
                new_ar[i,j] = re_eval(arg[i,j],replace_map)
        return new_ar

    # List
    elif (type(arg) is list or
          type(arg) is cvxpy_list):

        new_list = []
        for c in arg:
            new_list += [re_eval(c,replace_map)]
        return cvxpy_list(new_list)

    # Invalid
    else:
        raise TypeError('Invalid argument')



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

u"""
Noise weighting NTF synthesis (:mod:`pydsm.NTFdesign.weighting`)
================================================================

This modules provides code for the synthesis of the modulator NTF, based
on a noise weighting function.

.. currentmodule:: pydsm.NTFdesign.weighting


Functions
---------

.. autosummary::
   :toctree: generated/

    mult_weightings      -- Product of weighting functions
    q0_weighting         -- Q matrix from a noise weighting function
    ntf_fir_from_q0      -- design FIR NTF from matrix Q
    ntf_hybrid_weighting -- design hybrid NTF from weighting function
    ntf_fir_weighting    -- design FIR NTF from weighting function


Deprecated Functions
--------------------

.. autosummary::
   :toctree: generated/

   quantization_weighted_noise_gain    -- Alias of `quantization_noise_gain`
   q0_from_noise_weighting             -- Alias of `q0_weighting`
   synthesize_ntf_from_q0              -- Alias of `ntf_fir_from_q0`
   synthesize_ntf_from_noise_weighting -- Alias of `ntf_fir_weighting`
"""

from ._fir_weighting import *
from ._quantization_noise_gain import *

__all__ = (_fir_weighting.__all__ + _quantization_noise_gain.__all__
           + ['test'])

from ..._pytesttester import PytestTester
test = PytestTester(__name__)
del PytestTester



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..sets import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Quadratic form
def quad_form(x,P):
    r"""
    | :math:`\mbox{quad\_form}:
      \mathbb{R}^n \\times \mathbb{S}_+^n \\to \mathbb{R},
      \ \mbox{quad\_form}(x,P) = x^TPx`.
    | Convex.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param P: number or
              :ref:`matrix<matrix_obj>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check first input type
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        x = vstack([x])
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid first argument type')

    # Check shape
    (m,n) = x.shape
    if n!=1:
        raise ValueError('Invalid first argument dimension')

    # Check second input type
    if np.isscalar(P):
        P *= eye(m)
    elif type(P) is not cvxpy_matrix:
        raise TypeError('Invalid second argument type')

    # Check dimensions
    if P.shape[0] != m:
        raise ValueError('Invalid dimensions')

    # Check symmetry
    if not np.allclose(P,P.T):
        raise ValueError('Invalid second argument')

    # Check positive semidefinite
    if np.min(np.linalg.eig(P)[0]) < -EPSILON:
        raise ValueError('Invalid second argument')

    # Construct and return program
    z = variable(m,1)
    t = variable()
    p = program(minimize(t),
                [belongs(hstack((vstack((eye(m),z.T)),
                                 vstack((z,t)))),
                         semidefinite_cone)],
                [z],
                name='quad_form')
    return p(sqrtm(P)*x)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..interface import *
from ..arrays import cvxpy_array
from ..arrays import cvxpy_matrix

# min
def min(x):
    r"""
    | :math:`\mbox{min} :
      \mathbb{R}^{m \\times n} \\to \mathbb{R},
      \ \mbox{min}(X) = \mbox{min} \{ X_{ij} \ | \
      i \in [m], \ j \in [n] \}`.
    | Concave and nondecreasing.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`scalar object<scalar_ref>`.
    """

    # Check input
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        return x
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Single element
    if x.shape == (1,1):
        return x[0,0]

    # Construct program
    t = variable()
    z = variable(x.shape[0],x.shape[1])
    constr = []
    for i in range(0,x.shape[0],1):
        for j in range(0,x.shape[1],1):
            constr += [greater_equals(z[i,j],t)]
    p = program(maximize(t),
                constr,
                [z],
                name='min')

    # Return
    return p(x)



# -*- coding: utf-8 -*-

# Copyright (c) 2015, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

import numpy as np


__all__ = ["rms"]


def rms(x, no_dc=False):
    """
    Compute root means square value (RMS) of a vector.

    Given an input vector x, this function computes the square root of
    the sum of the squared elements of x, optionally removing the mean value
    of x from its entries beforehand.

    Parameters
    ----------
    x : array like
        input vector
    no_dc : bool, optional
        whether to remove the mean value, default is False

    Returns
    -------
    v : float
        RMS value

    Notes
    -----
    In formal terms, the operation being performed on :math:`n`-entry vector
    :math:`x` to return rms value :math:`v` is

    .. math::

        v = \\sqrt{\\frac{1}{n}\\sum_{i=0}^n x_i^2}

    Input x must be a 1 dimensional vector. If x is a matrix, the DELSIG
    version of rms returns a meaningless result, and so this function does...
    just a different result.
    """
    if no_dc:
        x = x - np.mean(x)
    return np.linalg.norm(x)/np.sqrt(len(x))



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

from ...correlations import raw_acorr

__all__ = ["q0_from_filter_ir"]


def q0_from_filter_ir(P, ir):
    """
    Legacy version of :func:`pydsm.NTFdesign.weighting.q0_weighting`

    Legacy method based on correlation for the computation of the
    description matrix for the weighting function used in FIR NTF design.

    Parameters
    ----------
    P : int
        order of the FIR to be eventually synthesized
    ir : array_like
        output filter description impulse response

    Returns
    -------
    q0 : ndarray
        the first row of the matrix Q used in the NTF optimization

    Notes
    -----
    The Q matrix being synthesized has (P+1) times (P+1) entries.

    See Also
    --------
    scipy.integrate.quad : integrator used internally.
        For the meaning of the integrator parameters.
    """
    # Manage optional parameters
    return raw_acorr(ir, P)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..sets import *
from ..utils import *
from ..interface import *
from ..arrays import cvxpy_matrix
from .geo_mean import geo_mean

# det_rootn
def det_rootn(x):
    r"""
    | :math:`\mbox{det\_rootn} :
      \mathbb{S}_+^n \\to \mathbb{R},
      \ \mbox{det\_rootn}(X) = (\mbox{det}(X))^{1/n}`.
    | Concave.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check type
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        x = vstack([x])
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Check shape
    (m,n) = x.shape
    if m!=n:
        raise ValueError('Invalid argument dimensions')

    # Construct objective and constraints
    v = variable(m,m)
    Z = variable(m,m,'lower_triangular')
    D = diag(Z)

    constr = [belongs(vstack((hstack((diag(D),Z.T)),
                              hstack((Z,v)))),
                      semidefinite_cone),geq(D,0)]

    # Add constraints if argument is not numeric
    if type(x).__name__ in ARRAY_OBJS:
        constr += [belongs(vstack([v]),semidefinite_cone),
                   equals(v,v.T)]

    # Check properties if argument is nunmeric
    elif (not np.allclose(x,x.T) or
          np.min(np.linalg.eig(x)[0]) < 0):
        return -np.inf

    # Construct program
    obj = geo_mean(D)
    p = program(maximize(obj),
                constr,
                [v],
                name='det_rootn')

    # Return program
    return p(x)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..sets import *
from ..interface import *
from ..arrays import cvxpy_array
from ..arrays import cvxpy_matrix

# log_sum_exp
def log_sum_exp(x):
    r"""
    | :math:`\mbox{log\_sum\_exp} :
      \mathbb{R}^n \\to \mathbb{R},
      \ \mbox{log\_sum\_exp}(x) =
      \mbox{log} \\big( \sum_{i = 1}^n e^{x_i} \\big)`.
    | Convex and increasing.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`scalar object<scalar_ref>`.
    """

    # Check input
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        return x
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Must be column
    if x.shape[1] != 1:
        raise ValueError('Invalid argument dimensions')

    # Single element
    if x.shape == (1,1):
        return x[0,0]

    # Construct program
    t = variable()
    z = variable(x.shape[0],1)
    v = variable(x.shape[0],1)
    w = variable(x.shape[0],1)
    constr = [equals(sum(w),1)]
    for i in range(0,x.shape[0],1):
        constr += [belongs(vstack((v[i,0]-t,1.,w[i,0])),
                           exp_cone),
                   less_equals(z[i,0],v[i,0])]
    p = program(minimize(t),
                constr,
                [z],
                name='log_sum_exp')

    # Return
    return p(x)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
import cvxopt as opt
from ..defs import *
from ..scalars import cvxpy_obj

# Class definition
class cvxpy_second_order_cone(object):
    r"""
    | :math:`\{(x,y) \in \mathbb{R}^n \\times \mathbb{R} \ |
      \ \|x\|_2 \le y \}`.
    """

    # Method: __init__
    def __init__(self):
        """
        Class constructor.
        """

        self.type = SET
        self.name = 'second_order_cone'
        self.expansion_type = SOC

    # Method: valid_shape
    def valid_shape(self,shape):

        return shape[0] > 1 == shape[1]

    # Method: __str__
    def __str__(self):

        return self.name

    # Method: _construct
    def _construct(self,el,mp,n):

        m = el.shape[0]-1
        G = opt.spmatrix(0.0,[],[],(m+1,n))
        h = opt.matrix(0.0,(m+1,1))

        # y
        y = el[m,0]
        if np.isscalar(y):
            h[0,0] = y*1.
        elif type(y) is cvxpy_obj:
            h[0,0] = y.value*1.
        else:
            G[0,mp[y]]=-1.

        # x
        for i in range(0,m,1):
            x = el[i,0]
            if np.isscalar(x):
                h[i+1,0] = x*1.
            elif type(x) is cvxpy_obj:
                h[i+1,0] = x.value*1.
            else:
                G[i+1,mp[x]] = -1.

        # Return G,h
        return G,h,m+1

# Create instance
second_order_cone = cvxpy_second_order_cone()



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#
from .exp_cone import exp_cone, cvxpy_exp_cone
from .kl_div_epi import kl_div_epi, cvxpy_kl_div_epi
from .geo_mean_cone import geo_mean_cone, cvxpy_geo_mean_cone
from .semidefinite_cone import semidefinite_cone, cvxpy_semidefinite_cone
from .second_order_cone import second_order_cone, cvxpy_second_order_cone
from .log_norm_cdf_hypo import log_norm_cdf_hypo, cvxpy_log_norm_cdf_hypo
from .power_pos_epi import power_pos_epi, cvxpy_power_pos_epi



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
import warnings
import importlib_resources
import pytest

try:
    import pytest_benchmark
    BENCHMARK_AVAILABLE = True
except ImportError:
    BENCHMARK_AVAILABLE = False

@pytest.mark.skipif(not BENCHMARK_AVAILABLE,
                    reason="pytest-benchmark is not installed")
@pytest.mark.benchmark(group="simulator")
class Benchmark_simulateDSM:

    @classmethod
    def setup_class(cls):
        # Load expected results
        with (importlib_resources.files('pydsm.delsig')
              .joinpath('tests/Data/test_simulateDSM_0.npz')
              .open('rb')) as f:
            cls.result = np.load(f)['arr_0']
        cls.N = cls.result.size
        # Take H as in H = synthesizeNTF(5, 32, 1)
        cls.H = (np.array([0.99604531+0.08884669j,  0.99604531-0.08884669j,
                           0.99860302+0.05283948j,  0.99860302-0.05283948j,
                           1.00000000+0.j]),
                 np.array([0.80655696+0.11982271j,  0.80655696-0.11982271j,
                           0.89807098+0.21981939j,  0.89807098-0.21981939j,
                           0.77776708+0.j]),
                 1)
        # Test frequency
        cls.f = 85.
        # Test input signal
        cls.u = 0.5*np.sin(2.*np.pi*cls.f/cls.N*np.arange(cls.N))

    def benchmark_simulateDSM_scipy_blas(self, benchmark):
        """Benchmark function for the scipy blas version of simulateDSM"""
        from pydsm.delsig._simulateDSM_scipy_blas import (
            simulateDSM as simulateDSM_scipyblas)
        output, da1, da2, da3 = benchmark(simulateDSM_scipyblas,
                                          self.u, self.H)
        np.testing.assert_equal(output, self.result)

    def benchmark_simulateDSM_cblas_blas(self, benchmark):
        """Benchmark function for the cblas version of simulateDSM"""
        try:
            from pydsm.delsig._simulateDSM_cblas import (
                simulateDSM as simulateDSM_cblas)
        except:
            pytest.skip("Cblas libraries not available")
        output, da1, da2, da3 = benchmark(simulateDSM_cblas,
                                          self.u, self.H)
        np.testing.assert_equal(output, self.result)

    @pytest.mark.slow
    def benchmark_simulateDSM_scipy(self, benchmark):
        """Benchmark function for the scipy version of simulateDSM"""
        from pydsm.delsig._simulateDSM_scipy import (
            simulateDSM as simulateDSM_scipy)
        from pydsm.exceptions import PyDsmSlowPathWarning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore",
                category=PyDsmSlowPathWarning)
            output, da1, da2, da3 = benchmark(simulateDSM_scipy,
                                              self.u, self.H)
        np.testing.assert_equal(output, self.result)



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
import importlib_resources
from pydsm.delsig import simulateDSM

__all__ = ["TestSimulateDSM"]


class TestSimulateDSM:

    def setUp(self):
        pass

    def test_default(self):
        with (importlib_resources.files('pydsm.delsig')
              .joinpath('tests/Data/test_simulateDSM_0.npz')
              .open('rb')) as f:
            d = np.load(f)['arr_0']
        # Take H as in H = synthesizeNTF(5, 32, 1)
        H = (np.array([0.99604531+0.08884669j,  0.99604531-0.08884669j,
                       0.99860302+0.05283948j,  0.99860302-0.05283948j,
                       1.00000000+0.j]),
             np.array([0.80655696+0.11982271j,  0.80655696-0.11982271j,
                       0.89807098+0.21981939j,  0.89807098-0.21981939j,
                       0.77776708+0.j]),
             1)
        N = 8192
        f = 85
        u = 0.5*np.sin(2.*np.pi*f/N*np.arange(N))
        v, d1, d2, d3 = simulateDSM(u, H)
        np.testing.assert_equal(v, d)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..sets import *
from ..interface import *
from ..arrays import cvxpy_matrix
from ..arrays import cvxpy_array

# log_norm_cdf
def log_norm_cdf(x):
    r"""
    | :math:`\mbox{log\_norm\_cdf} :
      \mathbb{R}^{m \\times n} \\to \mathbb{R}^{m \\times n},
      \ \mbox{log\_norm\_cdf}(X)_{ij} =
      \mbox{log} \\big(
      \int_{-\infty}^{X_{ij}} \\frac{1}{\sqrt{2\pi}}
      e^{-\\frac{1}{2}u^2} \mathrm{d}u \\big)`.
    | Concave and increasing.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number,
             :ref:`tree<tree_obj>`,
             :ref:`matrix<matrix_obj>` or
             :ref:`array<array_obj>`.
    """

    # Prepare input
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        arg = vstack([x])
    elif (type(x) is cvxpy_matrix or
          type(x).__name__ in ARRAY_OBJS):
        arg = x
    else:
        raise TypeError('Invalid argument')

    # Prepare output
    if type(arg) is cvxpy_matrix:
        output = zeros(arg.shape)
    else:
        output = cvxpy_array(arg.shape[0],arg.shape[1])

    # Construct program
    for i in range(0,arg.shape[0],1):
        for j in range(0,arg.shape[1],1):
            t = variable()
            v = variable()
            z = variable()
            p = program(maximize(t),
                        [belongs(vstack((v,t)),
                                 log_norm_cdf_hypo),
                         less_equals(v,z)],
                        [z],
                        name='log_norm_cdf')
            output[i,j] = p(arg[i,j])

    # Return output
    if output.shape == (1,1):
        return output[0,0]
    else:
        return output



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
from .scalars import cvxpy_obj
from .arrays import cvxpy_array
from .arrays import cvxpy_matrix
from functools import reduce

#***********************************************************************#
# Class definition: cvxpy_constr                                        #
#***********************************************************************#
class cvxpy_constr(object):

    # Method: __init__
    def __init__(self,left,constraint_type,right):
        """
        Class constructor.

        :param left: Left hand side.
        :param constraint type: Keyword (See cvxpy.defs).
        :param right: Right hand side.
        """

        self.left = left
        self.type = constraint_type
        self.right = right

    # Method: __getattribute__
    def __getattribute__(self,name):

        # Variables
        if name == 'variables':
            if self.type == BELONGS:
                return self.left.variables
            else:
                return cvxpy_list(set(self.left.variables +
                                      self.right.variables))

        # Parameters
        elif name == 'parameters':
            if self.type == BELONGS:
                return self.left.parameters
            else:
                return cvxpy_list(set(self.left.parameters +
                                      self.right.parameters))

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: __str__
    def __str__(self):

        l_text = str(self.left)
        type_text = self.type
        r_text = str(self.right)
        return l_text+' '+type_text+' '+r_text

    # Method: is_dcp
    def is_dcp(self):
        """
        Determines if the constraint is DCP-compliant.
        """

        if (self.type == EQUALS and
            self.left.is_affine() and
            self.right.is_affine()):
            return True
        elif (self.type == LESS_EQUALS and
              self.left.is_convex() and
              self.right.is_concave()):
            return True
        elif (self.type == GREATER_EQUALS and
              self.left.is_concave() and
              self.right.is_convex()):
            return True
        elif (self.type == BELONGS and
              self.left.is_affine()):
            return True
        else:
            return False

    # Method: is_affine
    def is_affine(self):

        if self.type == BELONGS:
            return False
        else:
            return self.left.is_affine() and self.right.is_affine()

#***********************************************************************#
# Class definition: cvxpy_list                                          #
#***********************************************************************#
class cvxpy_list(list):

    # Method: __getattributes__
    def __getattribute__(self,name):

        # Variables
        if name == 'variables':
            all_vars = reduce(lambda x,y: x+y,
                              list(map(lambda x: x.variables,self)),[])
            return cvxpy_list(set(all_vars))

        # Parameters
        elif name == 'parameters':
            all_params = reduce(lambda x,y: x+y,
                                list(map(lambda x: x.parameters,self)),[])
            return cvxpy_list(set(all_params))

        # Other
        else:
            return object.__getattribute__(self,name)

    # Method: _get_eq
    def _get_eq(self):
        """
        Gives equality constraints.
        """

        return cvxpy_list([c for c in self if c.type == EQUALS])

    # Method: _get_ineq_in
    def _get_ineq_in(self):
        """
        Gives inequality and membership constraints.
        """

        return cvxpy_list([c for c in self if c.type != EQUALS])

    # Method: is_dcp
    def is_dcp(self):
        """
        Determines if all constraints are DCP-compliant.
        """

        return all(list(map(lambda x:x.is_dcp(),self)))

    # Method: is_affine
    def is_affine(self):
        """
        Determines if all constraints are affine.
        """

        return all(list(map(lambda x:x.is_affine(),self)))

    # Method: __add__
    def __add__(self,other):

        return cvxpy_list(list(self) + other)

    # Method: __radd__
    def __radd__(self,other):

        return cvxpy_list(other + list(self))

    # Method: __str__
    def __str__(self):

        output = ''
        for i in range(0,len(self),1):
            output += str(self[i])
            if i != len(self)-1:
                output += '\n'
        return output

#***********************************************************************#
# Function definition: compare                                          #
#***********************************************************************#
def compare(obj1,constraint_type,obj2):
    """
    Compares obj1 with obj2.

    :param obj1: Left hand side obejct.
    :param constraint_type: Keyword (See cvxpy.defs.).
    :param obj2: Right hand side object.
    """

    # Both scalars
    if ((np.isscalar(obj1) or type(obj1).__name__ in SCALAR_OBJS) and
        (np.isscalar(obj2) or type(obj2).__name__ in SCALAR_OBJS)):

        # Upgrade scalars to cvxpy_obj
        if np.isscalar(obj1):
            obj1 = cvxpy_obj(CONSTANT,obj1,str(obj1))
        if np.isscalar(obj2):
            obj2 = cvxpy_obj(CONSTANT,obj2,str(obj2))

        # Construct and return constraint
        return cvxpy_constr(obj1,constraint_type,obj2)

    # Upgrate scalars to arrays
    if ((type(obj1) is cvxpy_matrix or type(obj1).__name__ in ARRAY_OBJS) and
        (np.isscalar(obj2) or type(obj2).__name__ in SCALAR_OBJS)):
        (m,n) = obj1.shape
        new_ar = cvxpy_array(m,n)
        for i in range(0,m,1):
            for j in range(0,n,1):
                new_ar[i,j] = obj2
        obj2 = new_ar
    if ((type(obj2) is cvxpy_matrix or type(obj2).__name__ in ARRAY_OBJS) and
        (np.isscalar(obj1) or type(obj1).__name__ in SCALAR_OBJS)):
        (m,n) = obj2.shape
        new_ar = cvxpy_array(m,n)
        for i in range(0,m,1):
            for j in range(0,n,1):
                new_ar[i,j] = obj1
        obj1 = new_ar

    # Both arrays
    if ((type(obj1) is cvxpy_matrix or type(obj1).__name__ in ARRAY_OBJS) and
        (type(obj2) is cvxpy_matrix or type(obj2).__name__ in ARRAY_OBJS)):
        constr = []
        if obj1.shape != obj2.shape:
            raise ValueError('Invalid dimensions')
        (m,n) = obj1.shape
        for i in range(0,m,1):
            for j in range(0,n,1):
                constr += [compare(obj1[i,j],constraint_type,obj2[i,j])]
        return cvxpy_list(constr)

    # Invalid arguments
    raise TypeError('Objects not comparable')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014–2025, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.


import sys
from setuptools import setup, Extension
from Cython.Build import cythonize
import platform
import numpy as np
import re

if sys.version_info[:2] < (3, 10):
    raise RuntimeError("Python 3 supported for versions >= 3.10")

link_prefix = r"https://github.com/sergiocallegari/PyDSM/blob/main/"

def to_abs_links(text):
    return re.sub(
        r"\]\((?!https?://)(.*?)\)",
        rf"]({link_prefix}\1)",
        text
    )

# Create readme file for PyPI
with open("README.md", encoding="utf-8") as f:
    readme_text = f.read()
readme_pypi_text = to_abs_links(readme_text)
with open("README-PyPI.md", "w", encoding="utf-8") as f:
    f.write(readme_pypi_text)

# Prepare the extension modules
ext_modules = cythonize([
    Extension(name='pydsm.delsig._simulateDSM_cblas',
              sources=['src/pydsm/delsig/_simulateDSM_cblas.pyx'],
              include_dirs=[np.get_include()]+["/usr/include/openblas/"],
              libraries=['blas'],
              define_macros=[('NPY_NO_DEPRECATED_API',
                              'NPY_1_7_API_VERSION')]),
    Extension(name='pydsm.delsig._simulateDSM_scipy_blas',
              sources=['src/pydsm/delsig/_simulateDSM_scipy_blas.pyx'],
              include_dirs=[np.get_include()],
              define_macros=[('NPY_NO_DEPRECATED_API',
                              'NPY_1_7_API_VERSION')])],
    compiler_directives={'language_level' : "3"})

# Special requirements for the windows platform
if platform.system() != 'Linux':
    # The cblas simulator is only built in Linux
    ext_modules = [
        Extension(name='pydsm.delsig._simulateDSM_scipy_blas',
                  sources=['src/pydsm/delsig/_simulateDSM_scipy_blas.pyx'],
                  include_dirs=[np.get_include()],
                  define_macros=[('NPY_NO_DEPRECATED_API',
                                  'NPY_1_7_API_VERSION')])]

setup(
    name='pydsm',
    setup_requires=["setuptools_scm"],
    ext_modules=ext_modules,
)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

from .abs import abs
from .det_rootn import det_rootn
from .exp import exp
from .geo_mean import geo_mean
from .huber import huber
from .kl_div import kl_div
from .lambda_max import lambda_max
from .lambda_min import lambda_min
from .log import log
from .log_norm_cdf import log_norm_cdf
from .log_sum_exp import log_sum_exp
from .max import max
from .min import min
from .norm1 import norm1
from .norm2 import norm2
from .norm_inf import norm_inf
from .nuclear_norm import nuclear_norm
from .power_abs import power_abs
from .power_p import power_p
from .power_pos import power_pos
from .quad_form import quad_form
from .quad_over_lin import quad_over_lin
from .sqrt import sqrt
from .square import square



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..sets import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Quadratic over linear
def quad_over_lin(x,y):
    r"""
    | :math:`\mbox{quad\_over\_lin} :
      \mathbb{R}^n \\times \mathbb{R}_{++} \\to \mathbb{R},
      \ \mbox{quad\_over\_lin}(x,y) = x^Tx/y`.
    | Convex in :math:`(x,y)` and decreasing in :math:`y`.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param y: number or
              :ref:`scalar object<scalar_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check x
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        x = vstack([x])
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid first argument')
    elif x.shape[1] != 1:
        raise TypeError('Invalid first argument')

    # Check y
    if (not np.isscalar(y) and
        type(y).__name__ not in SCALAR_OBJS):
        raise TypeError('Invalid second argument')

    # Check that y>0 if y is numeric
    if np.isscalar(y) and y <= 0:
        return np.inf

    # Construct objective and constraints
    n = x.shape[0]
    z = variable(n,1)
    w = variable()
    v = variable()
    t = variable()
    A = vstack((hstack((diag(v*ones((1,n))),z)),
                hstack((z.T,t))))
    constr = [belongs(A,semidefinite_cone),
              greater_equals(w,v)]

    # Construct and return program
    p = program(minimize(t),
                constr,
                [z,w],
                name='quad_over_lin')
    return p(x,y)



# -*- coding: utf-8 -*-

# Copyright (c) 2012, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import division, print_function

import numpy as np
from pydsm.delsig import clans
from pydsm.relab import cplxpair

__all__ = ["TestClans"]


class TestClans:

    def setUp(self):
        pass

    def test_clans(self):
        """Test function for clans"""
        e_z = [1,
               0.99604531 + 0.08884669j,
               0.99604531 - 0.08884669j,
               0.99860302 + 0.05283948j,
               0.99860302 - 0.05283948j]
        e_p = [0.41835234,
               0.48922229 + 0.1709716j,
               0.48922229 - 0.1709716j,
               0.65244884 + 0.38172239j,
               0.65244884 - 0.38172239j]
        e_k = 1
        e_z = cplxpair(e_z)
        e_p = cplxpair(e_p)
        z, p, k = clans(5, 32, 5, .95, 1)
        z = cplxpair(z)
        p = cplxpair(p)
        np.testing.assert_almost_equal(k, e_k, 6)
        np.testing.assert_almost_equal(z, e_z, 6)
        np.testing.assert_almost_equal(p, e_p, 6)

if __name__ == '__main__':
    run_module_suite()



# -*- coding: utf-8 -*-

# Copyright (c) 2015–2024, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import division

import numpy as np

from ._tf import evalTF


__all__ = ["rmsGain"]


def rmsGain(H, f1, f2, N=100):
    """
    Compute the root-mean-square gain of a DT transfer function.

    The computation is practiced on an assigned frequency interval.
    By root-mean-square, it is meant that the magnitude response of the
    transfer function is squared and averaged over the band of interest.
    Then, the square root of this averaged value is returned.

    Parameters
    ----------
    H : tuple
        transfer function either in (z,p,k) or (n,d) form
    f1 : real
        lower bound of frequency band on which the transfer function
        is evaluated
    f2 : real
        upper bound of frequency band on which the transfer function
        is evaluated
    N : int, optional
        number of points where the transfer function is evaluated in the
        interval. Defaults to 100.

    Returns
    -------
    rms : real
        rms value of the discrete time transfer function.

    Examples
    --------

    >>> import numpy as np
    >>> import pydsm
    >>> v = np.sin(np.linspace(0, 20*np.pi, 1024))
    >>> print(pydsm.delsig.rms(v))
    0.7067614298686651

    Notes
    -----
    The discrete-time transer function H is evaluated in the frequency band
    (f1,f2).  Spanning of the bandwidth is linear. Frequencies are normalized
    in the [0, 0.5] interval.

    Warning: the result of the computation is normalized in the number of
    points used for the computation, but not on the frequency range.

    This computation could be practiced much more accurately and possibly
    faster, using algorithms for the numerical computation of integrals.
    """
    w = np.linspace(2*np.pi*f1, 2*np.pi*f2, N)
    return np.linalg.norm(evalTF(H, np.exp(1j*w))) / np.sqrt(N)



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <https://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from .defs import *
import scipy.sparse as sp

import sys
if sys.version_info >= (3, 0):
    int_types = (int,)
else:
    int_types = (int, long)

#***********************************************************************#
# Class definition: cvxpy_array                                         #
#***********************************************************************#
class cvxpy_array(object):

    # Method: __init__
    def __init__(self,m,n):
        """
        Class constructor.

        :param m: Number of rows.
        :param n: Number of columns.
        """

        self.shape = (m,n)
        temp1 = {}
        for i in range(0,m,1):
            temp2 = {}
            for j in range(0,n,1):
                temp2[j] = 0.0
            temp1[i] = temp2
        self.data = temp1
        self.type = ARRAY

    # Method:__getattribute__
    def __getattribute__(self,name):

        # Transpose
        if name == 'T':
            new_ar = cvxpy_array(self.shape[1],self.shape[0])
            for i in range(0,self.shape[0],1):
                for j in range(0,self.shape[1],1):
                    new_ar.data[j][i] = self.data[i][j]
            return new_ar

        # Variables
        elif name == 'variables':
            l = []
            for i in range(0,self.shape[0],1):
                for j in range(0,self.shape[1],1):
                    if not np.isscalar(self.data[i][j]):
                        l += self.data[i][j].variables
            return cvxpy_list(set(l))

        # Parameters
        elif name == 'parameters':
            l = []
            for i in range(0,self.shape[0],1):
                for j in range(0,self.shape[1],1):
                    if not np.isscalar(self.data[i][j]):
                        l += self.data[i][j].parameters
            return cvxpy_list(set(l))

        # Value
        elif name == 'value':
            mat = cvxpy_matrix(np.zeros((self.shape[0],self.shape[1])),np.float64)
            for i in range(0,self.shape[0],1):
                for j in range(0,self.shape[1],1):
                    if np.isscalar(self.data[i][j]):
                        mat[i,j] = self.data[i][j]
                    else:
                        mat[i,j] = self.data[i][j].value
            return mat

        # Other attribute
        else:
            return object.__getattribute__(self,name)

    # Method: __setitem__
    def __setitem__(self,key,value):

        if type(key) != tuple:
            raise TypeError('Invalid Key')
        if len(key) != 2:
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[0]), int_types+(np.integer,))) or
            not (issubclass(type(key[1]), int_types+(np.integer,)))):
            raise TypeError('Invalid Key')
        if key[0] < 0 or key[0] >= self.shape[0]:
            raise ValueError('Index out of range')
        if key[1] < 0 or key[1] >= self.shape[1]:
            raise ValueError('Index out of range')

        self.data[key[0]][key[1]] = value

    # Method: __getitem__
    def __getitem__(self,key):

        if type(key) != tuple:
            raise TypeError('Invalid Key')
        if len(key) != 2:
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[0]), int_types+(np.integer,))) and
            type(key[0]) is not slice):
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[1]), int_types+(np.integer,))) and
            type(key[1]) is not slice):
            raise TypeError('Invalid Key')

        # Prepare left slice
        sl = [None,None,None]
        if np.isscalar(key[0]):
            sl = [key[0],key[0]+type(key[0])(1),1]
        else:
            if key[0].start == None:
                sl[0] = 0
            else:
                sl[0] = key[0].start
            if key[0].stop == None:
                sl[1] = self.shape[0]
            else:
                sl[1] = key[0].stop
            if key[0].step == None:
                sl[2] = 1
            else:
                sl[2] = key[0].step

        # Prepare right slices
        sr = [None,None,None]
        if np.isscalar(key[1]):
            sr = [key[1],key[1]+type(key[1])(1),1]
        else:
            if key[1].start == None:
                sr[0] = 0
            else:
                sr[0] = key[1].start
            if key[1].stop == None:
                sr[1] = self.shape[1]
            else:
                sr[1] = key[1].stop
            if key[1].step == None:
                sr[2] = 1
            else:
                sr[2] = key[1].step

        # Check type
        if not all(list(map(lambda x: (issubclass(type(x),int) or
                                       issubclass(type(x),np.integer)),
                            [sl[0],sl[1],sl[2],sr[0],sr[1],sr[2]]))):
            raise TypeError('Invalid key')

        # Calculate new size
        ll = range(sl[0],sl[1],sl[2])
        lr = range(sr[0],sr[1],sr[2])
        new_m = len(ll)
        new_n = len(lr)

        # Scalar
        if new_m == 1 and new_n == 1:
            if sl[0] < 0 or sl[0] >= self.shape[0]:
                raise ValueError('Index out of range')
            if sr[0] < 0 or sr[0] >= self.shape[1]:
                raise ValueError('Index out of range')
            return self.data[sl[0]][sr[0]]

        # Array
        new_ar = cvxpy_array(new_m,new_n)
        for i in range(0,new_m,1):

            if ll[i] < 0 or ll[i] >= self.shape[0]:
                raise ValueError('Index out of range')

            for j in range(0,new_n,1):

                if lr[j] < 0 or lr[j] >= self.shape[1]:
                    raise ValueError('Index out of range')

                new_ar.data[i][j] = self.data[ll[i]][lr[j]]
        return new_ar

    # Method: __str__
    def __str__(self):

        output = '['
        for i in range(0,self.shape[0],1):
            if i:
                output += '\n ['
            else:
                output += '['
            for j in range(0,self.shape[1],1):
                output += ' ' + str(self.data[i][j]) + ' '
            output += ']'
        output += '] '
        return output

    # Method: __add__
    def __add__(self,other):

        return self.__addsub__(other,LEFT_ADD)

    # Method: __sub__
    def __sub__(self,other):

        return self.__addsub__(other,LEFT_SUBTRACT)

    # Method: __radd__
    def __radd__(self,other):

        return self.__addsub__(other,RIGHT_ADD)

    # Method: __rsub__
    def __rsub__(self,other):

        return self.__addsub__(other,RIGHT_SUBTRACT)

    # Method: __addsub__
    def __addsub__(self,other,action):
        """
        Handles addition and subraction.

        :param other: Other operand.
        :param action: Keyword (See cvxpy.defs).
        """

        # New array
        (m,n) = self.shape
        new_ar = cvxpy_array(m,n)

        # Scalar
        if (np.isscalar(other) or
            type(other).__name__ in SCALAR_OBJS):
            for i in range(0,m,1):
                for j in range(0,n,1):
                    if action == LEFT_ADD:
                        new_ar.data[i][j] = self.data[i][j] + other
                    elif action == LEFT_SUBTRACT:
                        new_ar.data[i][j] = self.data[i][j] - other
                    elif action == RIGHT_ADD:
                        new_ar.data[i][j] = other + self.data[i][j]
                    else:
                        new_ar.data[i][j] = other - self.data[i][j]
            return new_ar

        # Array or matrix
        elif (type(other).__name__ in MATRIX_OBJS or
              type(other).__name__ in ARRAY_OBJS):
            if other.shape != self.shape:
                raise ValueError('Invalid Dimensions')
            else:
                for i in range(0,m,1):
                    for j in range(0,n,1):
                        if action == LEFT_ADD:
                            new_ar.data[i][j] = self.data[i][j] + other[i,j]
                        elif action == LEFT_SUBTRACT:
                            new_ar.data[i][j] = self.data[i][j] - other[i,j]
                        elif action == RIGHT_ADD:
                            new_ar.data[i][j] = other[i,j] + self.data[i][j]
                        else:
                            new_ar.data[i][j] = other[i,j] - self.data[i][j]
                return new_ar

        # Sparray or spmatrix
        elif (type(other) is cvxpy_spmatrix or
              type(other) is cvxpy_sparray):
            if other.shape != self.shape:
                raise ValueError('Invalid Dimensions')
            else:
                for i in range(0,m):
                    for j in range(0,n):
                        new_ar.data[i][j] = self.data[i][j]
                for i in range(0,m,1):
                    rowi_indeces = other.rows[i]
                    rowi_values = other.data[i]
                    for k in range(0,len(rowi_indeces)):
                        j = rowi_indeces[k]
                        value = rowi_values[k]
                        if action == LEFT_ADD:
                            new_ar.data[i][j] = new_ar.data[i][j] + value
                        elif action == LEFT_SUBTRACT:
                            new_ar.data[i][j] = new_ar.data[i][j] - value
                        elif action == RIGHT_ADD:
                            new_ar.data[i][j] = value + new_ar.data[i][j]
                        else:
                            new_ar.data[i][j] = value - new_ar.data[i][j]
                return new_ar

        # Not implemented
        else:
            return NotImplemented

    # Method: __mul__
    def __mul__(self,other):

        return self.__mulhandle__(other,LEFT_MULTIPLY)

    # Method: __rmul__
    def __rmul__(self,other):

        return self.__mulhandle__(other,RIGHT_MULTIPLY)

    # Method: __mulhandle__
    def __mulhandle__(self,other,action):
        """
        Handles multiplication.

        :param other: Other operand.
        :param action: Keyword (See cvxpy.defs).
        """

        # Dimensions
        (m,n) = self.shape

        # Scalar
        if (np.isscalar(other) or
            type(other).__name__ in SCALAR_OBJS):
            new_ar = cvxpy_array(m,n)
            for i in range(0,m):
                for j in range(0,n):
                    if action == LEFT_MULTIPLY:
                        new_ar.data[i][j] = self.data[i][j] * other
                    else:
                        new_ar.data[i][j] = other * self.data[i][j]
            return new_ar

        # Array or matrix
        elif (type(other).__name__ in MATRIX_OBJS or
              type(other).__name__ in ARRAY_OBJS):
            (p,q) = other.shape
            if action == LEFT_MULTIPLY:
                if p != n:
                    raise ValueError('Invalid dimensions')
                else:
                    new_ar = cvxpy_array(m,q)
                    for i in range(0,m):
                        for j in range(0,q):
                            temp = 0
                            for k in range(0,n):
                                temp += self.data[i][k]*other[k,j]
                            new_ar.data[i][j] = temp

                    # Convert to scalar if shape is (1,1)
                    if new_ar.shape == (1,1):
                        return new_ar.data[0][0]
                    else:
                        return new_ar
            else:
                if q != m:
                    raise ValueError('Invalid dimensions')
                else:
                    new_ar = cvxpy_array(p,n)
                    for i in range(0,p):
                        for j in range(0,n):
                            temp = 0
                            for k in range(0,m):
                                temp += other[i,k]*self.data[k][j]
                            new_ar.data[i][j] = temp

                    # Convert to scalar if shape = (1,1)
                    if new_ar.shape == (1,1):
                        return new_ar.data[0][0]
                    else:
                        return new_ar

        # Sparray or spmatrix
        elif (type(other) is cvxpy_spmatrix or
              type(other) is cvxpy_sparray):
            (p,q) = other.shape
            if action == LEFT_MULTIPLY:
                if p != n:
                    raise ValueError('Invalid dimensions')
                else:
                    new_ar = cvxpy_array(m,q)
                    other._create_col_based_rep()
                    for j in range(0,q):
                        colj_indeces = other.cols[j]
                        colj_values = other.datac[j]
                        for i in range(0,m):
                            temp = 0.
                            for r in range(0,len(colj_indeces)):
                                k = colj_indeces[r]
                                temp += self.data[i][k]*colj_values[r]
                            new_ar.data[i][j] = temp

                    # Convert to scalar if shape is (1,1)
                    if new_ar.shape == (1,1):
                        return new_ar.data[0][0]
                    else:
                        return new_ar
            else:
                if q != m:
                    raise ValueError('Invalid dimensions')
                else:
                    new_ar = cvxpy_array(p,n)
                    for i in range(0,p):
                        rowi_indeces = other.rows[i]
                        rowi_values = other.data[i]
                        for j in range(0,n):
                            temp = 0.
                            for r in range(0,len(rowi_indeces)):
                                k = rowi_indeces[r]
                                temp += rowi_values[r]*self.data[k][j]
                            new_ar.data[i][j] = temp

                    # Convert to scalar if shape = (1,1)
                    if new_ar.shape == (1,1):
                        return new_ar.data[0][0]
                    else:
                        return new_ar
        else:
            return NotImplemented

    # Method: __neg__
    def __neg__(self):

        # Allocate new array
        (m,n) = self.shape
        new_ar = cvxpy_array(m,n)

        # Fill and return array
        for i in range(0,m,1):
            for j in range(0,n,1):
                new_ar.data[i][j] = -self.data[i][j]
        return new_ar

    # Method: is_affine
    def is_affine(self):
        """
        Determines if the array contains affine
        expressions only.
        """

        for i in range(0,self.shape[0],1):
            for j in range(0,self.shape[1],1):
                if (not np.isscalar(self.data[i][j]) and
                    not self.data[i][j].is_affine()):
                    return False
        return True

#***********************************************************************#
# Class definition: cvxpy_var                                           #
#***********************************************************************#
class cvxpy_var(cvxpy_array):

    # Method: __init__
    def __init__(self,m,n,structure=None,name=None):
        """
        Class constructor.

        :param name: String.
        :param m: Number of rows.
        :param n: Number of columns.
        :param structure: Keyword (See cvxpy.defs).
        """

        # Set name
        if name == None:
            name = BASE_VAR+str(cvxpy_scalar_var.i)
            cvxpy_scalar_var.i += 1

        # Call parent constructor
        cvxpy_array.__init__(self,m,n)

        # No structure
        if structure == None:
            for i in range(0,m,1):
                for j in range(0,n,1):
                    v = cvxpy_scalar_var(name+'['+str(i)+','+str(j)+']')
                    self.data[i][j] = v

        # Lower triangular
        elif structure == LOWER_TRIANGULAR:
            if m!=n:
                raise ValueError('Invalid dimensions')
            for i in range(0,m,1):
                for j in range(0,i+1,1):
                    v = cvxpy_scalar_var(name+'['+str(i)+','+str(j)+']')
                    self.data[i][j] = v

        # Upper triangular
        elif structure == UPPER_TRIANGULAR:
            if m!=n:
                raise ValueError('Invalid dimensions')
            for i in range(0,m,1):
                for j in range(i,n,1):
                    v = cvxpy_scalar_var(name+'['+str(i)+','+str(j)+']')
                    self.data[i][j] = v

        # Symmetric
        elif structure == SYMMETRIC:
            if m!=n:
                raise ValueError('Invalid dimensions')
            for i in range(0,m,1):
                for j in range(0,i+1,1):
                    v = cvxpy_scalar_var(name+'['+str(i)+','+str(j)+']')
                    self.data[i][j] = v
                    self.data[j][i] = v

        # Error
        else:
            raise ValueError('Invalid structure')

#***********************************************************************#
# Class definition: cvxpy_param                                         #
#***********************************************************************#
class cvxpy_param(cvxpy_array):

    # Method: __init__
    def __init__(self,m,n,attribute=None,name=None):
        """
        Class constructor.

        :param name: String.
        :param m: Number of rows.
        :param n: Number of columns.
        :param attribute: Keyword (See cvxpy.defs).
        """

        # Set name
        if name == None:
            name = BASE_PARAM+str(cvxpy_scalar_param.i)
            cvxpy_scalar_param.i += 1

        # Call parent constructor
        cvxpy_array.__init__(self,m,n)

        # Fill in
        for i in range(0,m,1):
            for j in range(0,n,1):
                v = cvxpy_scalar_param(attribute,
                                       name+'['+str(i)+','+str(j)+']')
                self.data[i][j] = v

    # Method: __setattr__
    def __setattr__(self,name,value):

        # Value
        if name == 'value':

            # Check type
            if type(value) is not cvxpy_matrix:
                raise TypeError('Invalid value type')

            # Check dimensions
            if value.shape != self.shape:
                raise ValueError('Invalid value dimensions')

            # Store new values
            for i in range(0,self.shape[0],1):
                for j in range(0,self.shape[1],1):
                    self.data[i][j].value = value[i,j]

        # Other
        else:
            object.__setattr__(self,name,value)

#***********************************************************************#
# Class definition: cvxpy_matrix                                        #
#***********************************************************************#
class cvxpy_matrix(np.matrix):

    # Method: __init__
    def __init__(self,data,dtype=None,copy=True):
        """
        Class constructor.

        :param data: Array-like or string.
        :param dtype: Data type.
        :param copy: Boolean.

        See numpy.matrix documentation.
        """

        # Parent constructor is called automatically
        pass


    # Method:__getattr__
    def __getattribute__(self,name):

        # Inverse
        if name == 'I':
            temp = np.array(self.copy())
            temp = np.array(np.matrix(temp).I)
            return cvxpy_matrix(temp,np.dtype(np.float64))

        # Other attribute
        else:
            return np.matrix.__getattribute__(self,name)


    # Method: __add__
    def __add__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__add__(self,other)

    # Method: __radd__
    def __radd__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__radd__(self,other)

    # Method: __sub__
    def __sub__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__sub__(self,other)

    # Method: __rsub__
    def __rsub__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__rsub__(self,other)

    # Method: __mul__
    def __mul__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__mul__(self,other)

    # Method: __rmul__
    def __rmul__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.matrix.__rmul__(self,other)

#***********************************************************************#
# Class definition: cvxpy_spmatrix                                      #
#***********************************************************************#
class cvxpy_spmatrix(sp.lil_matrix):

    # Method: __init__
    def __init__(self,arg,dtype=None,copy=False):
        """
        Class constructor.

        :param arg: Dense matrix, sparse matrix or shape tuple.
        :param dtype: Data type.
        :param copy: Boolean.

        See scipy.sparse.lil_matrix documentation.
        """

        # Call parent constructor
        sp.lil_matrix.__init__(self,arg,dtype=dtype)
        self.format = sp.lil_matrix.__name__[:3]

    # Method: _create_col_based_rep
    def _create_col_based_rep(self):
        self.cols = {}
        self.datac = {}
        for j in range(0,self.shape[1]):
            self.cols[j] = []
            self.datac[j] = []
        for i in range(0,self.shape[0]):
            rowi_indeces = self.rows[i]
            rowi_values = self.data[i]
            for k in range(0,len(rowi_indeces)):
                j = rowi_indeces[k]
                value = rowi_values[k]
                self.cols[j].append(i)
                self.datac[j].append(value)

    # Method:__getattr__
    def __getattribute__(self,name):

        # Inverse
        if name == 'T':
            return cvxpy_spmatrix(self.transpose())

        # Other attribute
        else:
            return sp.lil_matrix.__getattribute__(self,name)

    # Method: __add__
    def __add__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return sp.lil_matrix.__add__(self,other)

    # Method: __radd__
    def __radd__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.lil_matrix.__radd__(self,other)

    # Method: __sub__
    def __sub__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.lil_matrix.__sub__(self,other)

    # Method: __rsub__
    def __rsub__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.lil_matrix.__rsub__(self,other)

    # Method: __mul__
    def __mul__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.lil_matrix.__mul__(self,other)

    # Method: __rmul__
    def __rmul__(self,other):

        if (type(other).__name__ in SCALAR_OBJS or
            type(other).__name__ in ARRAY_OBJS):
            return NotImplemented
        else:
            return np.lil_matrix.__rmul__(self,other)

#***********************************************************************#
# Class definition: cvxpy_sparray                                       #
#***********************************************************************#
class cvxpy_sparray(object):

    # Method: __init__
    def __init__(self,m,n):
        """
        Class constructor.

        :param m: Number of rows.
        :param n: Number of columns.
        """

        self.shape = (m,n)
        self.type = SPARRAY
        self.rows = {}
        self.data = {}
        for i in range(0,m):
            self.rows[i] = []
            self.data[i] = []
        self.nnz = 0

    # Method: _create_col_based_rep
    def _create_col_based_rep(self):
        self.cols = {}
        self.datac = {}
        for j in range(0,self.shape[1]):
            self.cols[j] = []
            self.datac[j] = []
        for i in range(0,self.shape[0]):
            rowi_indeces = self.rows[i]
            rowi_values = self.data[i]
            for k in range(0,len(rowi_indeces)):
                j = rowi_indeces[k]
                value = rowi_values[k]
                self.cols[j].append(i)
                self.datac[j].append(value)

    # Method:__getattribute__
    def __getattribute__(self,name):

        # Transpose
        if name == 'T':
            new_ar = cvxpy_sparray(self.shape[1],self.shape[0])
            for i in range(0,self.shape[0],1):
                rowi_indeces = self.rows[i]
                rowi_values = self.data[i]
                for k in range(0,len(rowi_indeces)):
                    new_ar[rowi_indeces[k],i] = rowi_values[k]
            return new_ar

        # Variables
        elif name == 'variables':
            l = []
            for i in range(0,self.shape[0],1):
                for obj in self.data[i]:
                    if not np.isscalar(obj):
                        l += obj.variables
            return cvxpy_list(set(l))

        # Parameters
        elif name == 'parameters':
            l = []
            for i in range(0,self.shape[0],1):
                for obj in self.data[i]:
                    if not np.isscalar(obj):
                        l += obj.parameters
            return cvxpy_list(set(l))

        # Value
        elif name == 'value':
            mat = cvxpy_spmatrix((self.shape[0],self.shape[1]),np.float64)
            for i in range(0,self.shape[0],1):
                rowi_indeces = self.rows[i]
                rowi_values = self.data[i]
                for k in range(0,len(rowi_indeces)):
                    if np.isscalar(rowi_values[k]):
                        mat[i,rowi_indeces[k]] = rowi_values[k]
                    else:
                        mat[i,rowi_indeces[k]] = rowi_values[k].value
            return mat

        # Other attribute
        else:
            return object.__getattribute__(self,name)

    # Method: __setitem__
    def __setitem__(self,key,value):

        if type(key) != tuple:
            raise TypeError('Invalid Key')
        if len(key) != 2:
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[0]),int) or
                 issubclass(type(key[0]),np.integer)) or
            not (issubclass(type(key[1]),int) or
                 issubclass(type(key[1]),np.integer))):
            raise TypeError('Invalid Key')
        if key[0] < 0 or key[0] >= self.shape[0]:
            raise ValueError('Index out of range')
        if key[1] < 0 or key[1] >= self.shape[1]:
            raise ValueError('Index out of range')

        i = key[0]
        j = key[1]
        rowi_indeces = self.rows[i]
        rowi_values = self.data[i]

        found = False
        k = 0
        while k < len(rowi_indeces):

            if j > rowi_indeces[k]:
                k += 1

            elif j == rowi_indeces[k]:
                found = True
                break;

            else:
                break

        # present
        if found:
            if value == 0.:
                self.rows[i] = rowi_indeces[:k]+rowi_indeces[k+1:]
                self.data[i] = rowi_values[:k]+rowi_values[k+1:]
                self.nnz -= 1
            else:
                rowi_values[k] = value

        # not present
        else:
            if value != 0.:
                rowi_indeces.insert(k,j)
                rowi_values.insert(k,value)
                self.nnz += 1

    # Method: __getitem__
    def __getitem__(self,key):

        if type(key) != tuple:
            raise TypeError('Invalid Key')
        if len(key) != 2:
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[0]),int) or
                 issubclass(type(key[0]),np.integer)) and
            type(key[0]) is not slice):
            raise TypeError('Invalid Key')
        if (not (issubclass(type(key[1]),int) or
                 issubclass(type(key[1]),np.integer)) and
            type(key[1]) is not slice):
            raise TypeError('Invalid Key')

        # Prepare left slice
        sl = [None,None,None]
        if np.isscalar(key[0]):
            sl = [key[0],key[0]+type(key[0])(1),1]
        else:
            if key[0].start == None:
                sl[0] = 0
            else:
                sl[0] = key[0].start
            if key[0].stop == None:
                sl[1] = self.shape[0]
            else:
                sl[1] = key[0].stop
            if key[0].step == None:
                sl[2] = 1
            else:
                sl[2] = key[0].step

        # Prepare right slices
        sr = [None,None,None]
        if np.isscalar(key[1]):
            sr = [key[1],key[1]+type(key[1])(1),1]
        else:
            if key[1].start == None:
                sr[0] = 0
            else:
                sr[0] = key[1].start
            if key[1].stop == None:
                sr[1] = self.shape[1]
            else:
                sr[1] = key[1].stop
            if key[1].step == None:
                sr[2] = 1
            else:
                sr[2] = key[1].step

        # Check type
        if not all(list(map(lambda x: (issubclass(type(x),int) or
                                       issubclass(type(x),np.integer)),
                            [sl[0],sl[1],sl[2],sr[0],sr[1],sr[2]]))):
            raise TypeError('Invalid key')

        # Calculate new size
        ll = range(sl[0],sl[1],sl[2])
        lr = range(sr[0],sr[1],sr[2])
        new_m = len(ll)
        new_n = len(lr)

        # Scalar
        if new_m == 1 and new_n == 1:
            if sl[0] < 0 or sl[0] >= self.shape[0]:
                raise ValueError('Index out of range')
            if sr[0] < 0 or sr[0] >= self.shape[1]:
                raise ValueError('Index out of range')

            i = sl[0]
            j = sr[0]
            rowi_indeces = self.rows[i]
            rowi_values = self.data[i]

            for k in range(0,len(rowi_indeces)):
                if j == rowi_indeces[k]:
                    return rowi_values[k]
            return np.float64(0.)

        # Array
        new_ar = cvxpy_sparray(new_m,new_n)
        for i in range(0,new_m,1):

            lli = ll[i]

            if lli < 0 or lli >= self.shape[0]:
                raise ValueError('Index out of range')

            rowi_indeces = self.rows[lli]
            rowi_values = self.data[lli]

            for j in range(0,new_n,1):

                lrj = lr[j]

                if lrj < 0 or lrj >= self.shape[1]:
                    raise ValueError('Index out of range')

                for k in range(0,len(rowi_indeces)):
                    if lrj == rowi_indeces[k]:
                        new_ar[i,j] = rowi_values[k]
        return new_ar

    # Method: __str__
    def __str__(self):

        output = ''
        for i in range(0,self.shape[0],1):
            rowi_indeces = self.rows[i]
            rowi_values = self.data[i]
            for k in range(0,len(rowi_indeces)):
                output += ' (' + str(i)+','+str(rowi_indeces[k])+')\t\t'
                output += str(rowi_values[k]) + '\n'
        return output[:-1]

# Load modules
from .scalars import cvxpy_scalar_var
from .scalars import cvxpy_scalar_param
from .constraints import cvxpy_list



# -*- coding: utf-8 -*-

# Copyright (c) 2012–2025, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <https://www.gnu.org/licenses/>.

# This file includes code ported from the DELSIG Matlab toolbox
# (see https://www.mathworks.com/matlabcentral/fileexchange/19)
# covered by the following copyright and permission notice
#
# Copyright (c) 2009 Richard Schreier
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the distribution
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Code ported/reimplemented from the DELSIG toolbox (:mod:`pydsm.delsig`)
=======================================================================

The DELSIG toolbox for MATLAB is copyrighted by R. Schreier and
distributed under the BSD license.

.. currentmodule:: pydsm.delsig


Key functions
-------------

.. autosummary::
   :toctree: generated/

   synthesizeNTF
   clans
   synthesizeChebyshevNTF
   simulateDSM

Other selected functions
------------------------

Delta sigma utilities
.....................

.. autosummary::
   :toctree: generated/

   partitionABCD
   rmsGain

General utilities
.................

.. autosummary::
   :toctree: generated/

   dbv
   dbp
   undbv
   undbp
   dbm
   undbm
   rms

Graphing
........

.. autosummary::
   :toctree: generated/

   plotPZ
   axisLabels


Plumbing
--------

.. autosummary::
   :toctree: generated/

   ds_synNTFobj1
   ds_f1f2
   ds_optzeros
   dsclansNTF
   padl
   padr
   padt
   padb
   evalTF
   evalRPoly

"""

__delsig_version__ = "7.4"

# The delsig module reflects the flat organization of the original DELSIG
from ._axisLabels import *
from ._decibel import *
from ._ds import *
from ._padding import *
from ._tf import *
from ._plot import *
from ._synthesizeNTF import *
from ._synthesizeChebyshevNTF import *
from ._clans import *
from ._dsclansNTF import *
from ._simulateDSM import *
from ._simulateDSM_scipy import *
from ._partitionABCD import *
from ._rmsGain import *
from ._rms import *

from .._pytesttester import PytestTester
test = PytestTester(__name__)
bench = PytestTester(__name__, mode='benchmarks')
del PytestTester
